from __future__ import division, print_function

import numpy as np
import mne
import pandas as pd
from scipy.fftpack import fft,fftfreq,ifft,fftshift
from scipy.ndimage import convolve1d
from scipy.signal import filter_design, resample, filtfilt, hann 
import matplotlib.pyplot as plt
import nitime.algorithms as ntalg
from sklearn.decomposition import PCA
from sklearn.linear_model import Ridge


def lowpass_filter(s, sample_rate, cutoff_freq, filter_order=5, rescale=False):
    """
        Lowpass filter a signal s, with sample rate sample_rate.

        s: the signal (n_channels x n_timepoints)
        sample_rate: the sample rate in Hz of the signal
        cutoff_freq: the cutoff frequency of the filter
        filter_order: the order of the filter...

        Returns the low-pass filtered signal s.
    """
    #create a butterworth filter
    nyq = sample_rate / 2.0
    b,a = filter_design.butter(filter_order, cutoff_freq / nyq)

    #filter the signal
    filtered_s = filtfilt(b, a, s)

    if rescale:
        #rescale filtered signal
        filtered_s /= filtered_s.max()
        filtered_s *= s.max()

    return filtered_s

def highpass_filter(s, sample_rate, cutoff_freq, filter_order=5, rescale=False):
    """
        Highpass filter a signal s, with sample rate sample_rate.

        s: the signal (n_channels x n_timepoints)
        sample_rate: the sample rate in Hz of the signal
        cutoff_freq: the cutoff frequency of the filter
        filter_order: the order of the filter...

        Returns the high-pass filtered signal s.
    """
    #create a butterworth filter
    nyq = sample_rate / 2.0
    b,a = filter_design.butter(filter_order, cutoff_freq / nyq, btype='high')

    #filter the signal
    filtered_s = filtfilt(b, a, s)

    if rescale:
        #rescale filtered signal
        filtered_s /= filtered_s.max()
        filtered_s *= s.max()

    return filtered_s


def bandpass_filter(s, sample_rate, low_freq, high_freq, filter_order=5, rescale=False):
    """
        Bandpass filter a signal s.

        s: the signal (n_channels x n_timepoints)
        sample_rate: the sample rate in Hz of the signal
        low_freq: the lower cutoff frequency
        upper_freq: the upper cutoff frequency
        filter_order: the order of the filter...

        Returns the bandpass filtered signal s.
    """
    #create a butterworth filter
    nyq = sample_rate / 2.0
    f = np.array([low_freq, high_freq]) / nyq
    b,a = filter_design.butter(filter_order, f, btype='bandpass')

    #filter the signal
    filtered_s = filtfilt(b, a, s)

    if rescale:
        #rescale filtered signal
        filtered_s /= filtered_s.max()
        filtered_s *= s.max()

    return filtered_s


def resample_signal(s, sample_rate, desired_sample_rate):
    """
        Resamples a signal from sample rate to desired_sample_rate.

        s: the signal
        sample_rate: the sample rate of the signal
        desired_sample_rate: the desired sample rate

        Returns t_rs,rs where t_rs is the time corresponding to each resampled point, rs is the resampled sigal.
    """

    duration = float(len(s)) / sample_rate
    t = np.arange(len(s)) * (1.0 / sample_rate)
    desired_n = int(duration*desired_sample_rate)
    rs,t_rs = resample(s, desired_n, t=t)
    return t_rs,rs


def power_spectrum(s, sr, log=False, max_val=None, hanning=False):

    sw = s
    if hanning:
        sw = s*np.hanning(len(s))

    f = fft(sw)
    freq = fftfreq(len(sw), d=1.0/sr)
    findex = freq >= 0.0
    ps = np.abs(f)**2
    if log:
        if max_val is None:
            max_val = ps.max()
        ps /= max_val
        ps = 20.0*np.log10(ps)

    return freq[findex], ps[findex]


def mt_power_spectrum(s, sample_rate, window_size, low_bias=False, bandwidth=5.0):
    """
        Computes a jackknifed multi-taper power spectrum of a given signal. The jackknife is over
        windowed segments of the signal, specified by window_size.
    """

    sample_length_bins = min(len(s), int(window_size * sample_rate))

    #break signal into chunks and estimate coherence for each chunk
    nchunks = int(np.floor(len(s) / float(sample_length_bins)))
    nleft = len(s) % sample_length_bins
    #ignore the last chunk if it's too short
    if nleft > (sample_length_bins / 2.0):
        nchunks += 1

    ps_freq = None
    ps_ests = list()
    for k in range(nchunks):
        si = k*sample_length_bins
        ei = min(len(s), si + sample_length_bins)
        print('si=%d, ei=%d, len(s)=%d' % (si, ei, len(s)))

        ps_freq,mt_ps,var = ntalg.multi_taper_psd(s[si:ei], Fs=sample_rate, adaptive=True, BW=bandwidth, jackknife=False,
                                                  low_bias=low_bias, sides='onesided')
        ps_ests.append(mt_ps)

    ps_ests = np.array(ps_ests)

    ps_mean = ps_ests.mean(axis=0)
    ps_std = ps_ests.std(axis=0, ddof=1)

    return ps_freq,ps_mean,ps_std


def match_power_spectrum(s, sample_rate, nsamps=5, isreal=False):
    """
        Create a signals that have the same power spectrum as s but with randomly shuffled phases. nsamps is the number
        of times the signal is permuted. Returns an nsamps X len(s) matrix.
    """

    #get FT of the signal
    sfft = fft(s)
    amplitude = np.abs(sfft)
    phase = np.angle(sfft)

    s_recon = np.zeros([nsamps, len(s)], dtype='complex128')
    for k in range(nsamps):
        #shuffle the phase
        np.random.shuffle(phase)
        #reconstruct the signal
        sfft_recon = amplitude*(np.cos(phase) + 1j*np.sin(phase))
        s_recon[k, :] = ifft(sfft_recon)

    if isreal:
        return np.real(s_recon)
    return s_recon


def gaussian_window(N, nstd):
    """
        Generate a Gaussian window of length N and standard deviation nstd.
    """
    hnwinlen = (N + (1-N%2)) // 2
    gauss_t = np.arange(-hnwinlen, hnwinlen+1, 1.0)
    gauss_std = float(N) / float(nstd)
    gauss_window = np.exp(-gauss_t**2 / (2.0*gauss_std**2)) / (gauss_std*np.sqrt(2*np.pi))
    return gauss_t,gauss_window


def find_extrema(s):
    """
        Find the max and mins of a signal s.
    """
    max_env = np.logical_and(
                        np.r_[True, s[1:] > s[:-1]],
                        np.r_[s[:-1] > s[1:], True])
    min_env = np.logical_and(
                        np.r_[True, s[1:] < s[:-1]],
                        np.r_[s[:-1] < s[1:], True])
    max_env[0] = max_env[-1] = False

    #exclude endpoints
    mini = [m for m in min_env.nonzero()[0] if m != 0 and m != len(s)-1]
    maxi = [m for m in max_env.nonzero()[0] if m != 0 and m != len(s)-1]

    return mini,maxi


def analytic_signal(s):
    """
        An implementation of computing the analytic signal.
    """

    #take FFT
    sfft = fft(s)
    freq = fftfreq(len(s))

    #zero out coefficients at negative frequencies
    sfft[freq < 0.0] = np.zeros(np.sum(freq < 0.0), dtype='complex')
    sfft[freq >= 0.0] *= 2.0

    #take the IFFT
    z = ifft(sfft)
    return z


def compute_instantaneous_frequency(z, sample_rate):
    """
        Compute the instantaneous frequency given an analytic signal z.
    """
    x = z.real
    y = z.imag

    dx = np.r_[0.0, np.diff(x)]*sample_rate
    dy = np.r_[0.0, np.diff(y)]*sample_rate

    f = (x*dy - y*dx) / (2*np.pi*(x**2 + y**2))

    return f


def demodulate(Z, over_space=True, depth=1):
    """
        Apply demodulation (Argawal et. al 2014) to a matrix of complex-valued signals Z.

        Args:
            Z: an NxT signal matrix of N complex valued signals, each of length T
            over_space: whether to demodulate across space (does PCA on N dimensions) or time (does PCA on T dimensions)
            depth: how many PCA projection phases to subtract off

        Returns:
            phase: An NxT real-valued matrix of demodulated phases.
            pcs: An NxN complex-valued matrix of principle components.
    """

    #do complex PCA on each IMF
    N,T = Z.shape

    if over_space:

        #construct a matrix with the real and imaginary parts separated
        X = np.zeros([2*N, T], dtype='float')
        X[:N, :] = Z.real
        X[N:, :] = Z.imag

        pca = PCA()
        pca.fit(X.T)

        complex_pcs = np.zeros([N, N], dtype='complex')
        for j in range(N):
            pc = pca.components_[j, :]
            complex_pcs[j, :].real = pc[:N]
            complex_pcs[j, :].imag = pc[N:]

        phase = np.angle(Z)
        for k in range(depth):
            #compute the kth PC projected component
            proj = np.dot(Z.T.squeeze(), complex_pcs[k, :].squeeze())
            phase -= np.angle(proj)

    else:

        first_pc = np.zeros([T], dtype='complex')

        pca_real = PCA(n_components=1, svd_solver="randomized")
        pca_real.fit(Z.real)
        print('pca_real.components_.shape=',pca_real.components_.shape)
        first_pc.real = pca_real.components_.squeeze()
        
        pca_imag = PCA(n_components=1, svd_solver="randomized")
        pca_imag.fit(Z.imag)
        print('pca_imag.components_.shape=',pca_imag.components_.shape)
        first_pc.imag = pca_imag.components_.squeeze()

        complex_pcs = np.array([first_pc])

        proj = first_pc

        #demodulate the signal
        phase = np.angle(Z) - np.angle(proj)

    return phase,complex_pcs


def compute_coherence_over_time(signal, trials, Fs, n_perm=5, low=0, high=300):
    '''
    Computes the coherence between the mean of subsets of trails. This can be used
    to assess signal stability in response to a stimulus (repeated or otherwise).

    INPUTS
    --------
    signal : array-like
        The array of neural signals. Should be time x signals

    trials : pd.DataFrame, contains columns 'epoch', and 'time'
             and same first dimension as signal
        A dataframe with time indices and trial number within each epoch (trial)
        This is used to pull out the corresponding timepoints from signal.

    Fs : int
        The sampling rate of the signal

    OUTPUTS
    --------
    coh_perm : np.array, shape (n_perms, n_signals, n_freqs)
        A collection of coherence values for each permutation.

    coh_freqs : np.array, shape (n_freqs)
        The frequency values corresponding to the final dimension of coh_perm
    Output is permutations x signals x frequency bands
    '''
    trials = pd.DataFrame(trials)
    assert ('epoch' in trials.columns and 'time' in trials.columns), 'trials must be a DataFrame with "epoch" column'
    n_trials = np.max(trials['epoch'])

    coh_perm = []
    for perm in xrange(n_perm):
        trial_ixs = np.random.permutation(np.arange(n_trials))
        t1 = trial_ixs[:n_trials//2]
        t2 = trial_ixs[n_trials//2:]
        
        # Split up trials and take the mean of each
        mn1, mn2 = [signal[trials.eval('epoch in @t_ix and time > 0').values].mean(level=('time'))
                    for t_ix in [t1, t2]]

        # Now compute coherence between the two
        coh_all_freqs = []
        for (elec, vals1), (_, vals2) in zip(mn1.iteritems(), mn2.iteritems()):
            ts_arr = np.vstack([vals1, vals2])
            coh, coh_freqs, coh_times, n_epochs, n_tapers = mne.connectivity.spectral_connectivity(ts_arr[None, :, :], sfreq=Fs,
                                                                                                   fmin=low, fmax=high, verbose=0)
            coh_all_freqs.append(coh[1, 0, :])
        coh_perm.append(coh_all_freqs)
    coh_perm = np.array(coh_perm)
    return coh_perm, coh_freqs


def break_envelope_into_events(s, threshold=0, merge_thresh=None, max_amp_thresh=None):
    """ Segments a one dimensional positive-valued time series into events with start and end times.

    :param s: The signal, a numpy array.
    :param threshold: The threshold for determining the onset of an event. When the amplitude of s
            exceeds threshold, an event starts, and when the amplitude of the signal falls below
            threshold, the event ends.
    :param merge_thresh: Events that are separated by less than minimum_len get merged together. minimum_len
            must be specified in number of time points, not actual time.
    :param max_amp_thresh: If not None, events whose maximum amplitude is below max_amp_thresh are discarded.

    :return: A list of event start and end times, and the maximum amplitude encountered in that event.
    """

    assert np.sum(s < 0) == 0, "segment_envelope: Can't segment a signal that has negative values!"

    # array to keep track of start and end times of each event
    events = list()

    # scan through the signal, find events
    in_event = False
    max_amp = -np.inf
    start_index = -1
    for t,x in enumerate(s):

        if in_event:
            if x > max_amp:
                # we found a new peak
                max_amp = x
            if x <= threshold:
                # the event has ended
                in_event = False
                events.append( (start_index, t, max_amp))
                start_index = -1
                #print 'Identified event (%d, %d, %0.6f)' % (start_index, t, max_amp)
        else:
            if x > threshold:
                in_event = True
                start_index = t
                max_amp = threshold

    # get the last event if there is one
    if start_index != -1:
        events.append( (start_index, len(s)-1, max_amp))

    # print '# of events (pre-merge): %d' % len(events)
    events = np.array(events)

    # discard any events whose maximum amplitude is less than max_amp_thresh
    if max_amp_thresh is not None:
        events2 = list()
        for si,ei,max_amp in events:
            if max_amp >= max_amp_thresh:
                events2.append( (si, ei, max_amp))
        events = np.array(events2)
        del events2

    if merge_thresh is None:
        return events

    #compute the inter-event interval, used for merging smaller events into larger ones
    iei = events[1:, 0] - events[:-1, 1]

    #create an empty list for merged events
    merged_events = list()

    #set the "current event" to be the first event
    estart, eend, eamp = events[0, :]

    for k in range(len(events)-1):

        #get the event at time k+1
        stime,etime,amp = events[k+1, :]

        #get the inter-event-interval between the event at time k+1 and k
        the_iei = iei[k]
        #print 'k=%d, the_iei=%d, merge_thresh=%d' % (k, the_iei, merge_thresh)

        if the_iei < merge_thresh:
            #extend the end time of the current event
            eend = etime
            #change the maximum peak of the current event
            eamp = max(eamp, amp)
        else:
            #don't merge, pop the previous event
            merged_events.append( (estart, eend, eamp))

            #set the currente event to be the event at k+1
            estart = stime
            eend = etime
            eamp = amp

    #pop the last event
    merged_events.append( (estart, eend, eamp))

    # print '# of merged events: %d' % len(merged_events)

    return np.array(merged_events)


def power_amplifier(s, thresh, pwr=2):
    """ Amplify elements of a positive-valued signal. Rescale the signal
        so that elements above thresh are equal to or greater than 1,
        and elements below thresh are less than one. Then take a power
        of the signal, which will supress values less than 1, and amplify
        values that are greater than one.
    """

    #normalize the signal
    s /= s.max()

    #shift the signal so elements at the threshold are set to 1
    s += 1.0 - thresh

    #raise the envelope to a power, amplifying values that are above 1
    s = s**pwr

    #re-normalize
    s -= (1.0 - thresh)**pwr
    s /= s.max()

    return s


def phase_locking_value(z1, z2):
    """ Compute the phase-locking-value (PLV) between two complex signals. """

    assert len(z1) == len(z2), "Signals must be same length! len(z1)=%d, len(z2)=%d" % (len(z1), len(z2))
    N = len(z1)
    theta = np.angle(z2) - np.angle(z1)

    p = np.exp(complex(0, 1)*theta)
    plv = np.abs(p.sum()) / N

    return plv


def correlation_function(s1, s2, lags, mean_subtract=True, normalize=True):
    """ Computes the cross-correlation function between signals s1 and s2. The cross correlation function is defined as:

            cf(k) = sum_over_t( (s1(t) - s1.mean()) * (s2(t+k) - s2.mean()) ) / s1.std()*s2.std()

    :param s1: The first signal.
    :param s2: The second signal.
    :param lags: An array of integers indicating the lags. The lags are in units of sample period.
    :param mean_subtract: If True, subtract the mean of s1 from s1, and the mean of s2 from s2, which is the standard thing to do.
    :param normalize: If True, then divide the correlation function by the product of standard deviations of s1 and s2.
    :return: cf The cross correlation function evaluated at the lags.
    """
    
    assert len(s1) == len(s2), "Signals must be same length! len(s1)=%d, len(s2)=%d" % (len(s1), len(s2))
    assert np.sum(np.isnan(s1)) == 0, "There are NaNs in s1"
    assert np.sum(np.isnan(s2)) == 0, "There are NaNs in s2"

    s1_mean = 0
    s2_mean = 0
    if mean_subtract:
        s1_mean = s1.mean()
        s2_mean = s2.mean()

    s1_std = s1.std(ddof=1)
    s2_std = s2.std(ddof=1)
    s1_centered = s1 - s1_mean
    s2_centered = s2 - s2_mean
    N = len(s1)

    assert N > lags.max(), "Lags are too long, length of signal is %d, lags.max()=%d" % (N, lags.max())

    cf = np.zeros([len(lags)])
    for k,lag in enumerate(lags):

        if lag == 0:
            cf[k] = np.dot(s1_centered, s2_centered) / N
        elif lag > 0:
            cf[k] = np.dot(s1_centered[:-lag], s2_centered[lag:]) / (N-lag)
            """
            if np.isnan(cf[k]):
                print 's1_centered=',s1_centered[:-lag]
                print 's2_centered=',s2_centered[lag:]
                plt.figure()
                plt.plot(s1_centered[:-lag], 'r-')
                plt.plot(s2_centered[lag:], 'b-')
                plt.legend(['s1', 's2'])
                plt.axis('tight')
                print 'There is a nan, lag=%d, k=%d, N=%d, len(s1_centered)=%d, len(s2_centered)=%d...' % (lag, k, N, len(s1_centered), len(s2_centered))
                plt.show()
            """

        elif lag < 0:
            cf[k] = np.dot(s1_centered[np.abs(lag):], s2_centered[:lag]) / (N+lag)

    if normalize:
        cf /= s1_std * s2_std

    return cf

def linear_filter1D(sin, sout, lag=0, debug=0):
    """ Estimates the linear filter  between sin and sout which are both one dimensional arrays of
    equal length. Estimation based on the normal equation in the Fourier Domain.
    lags is the number of points in the past of the filter.
    signals are zeroed but the bias term is not returned.
    returns the weights of the filter."""
    
    assert len(sin) == len(sout), "Signals must be same length! len(sin)=%d, len(sout)=%d" % (len(sin), len(sout))
    assert np.sum(np.isnan(sin)) == 0, "There are NaNs in sin"
    assert np.sum(np.isnan(sout)) == 0, "There are NaNs in sout"
    
    lags = np.asarray(range(-lag, lag+1, 1))             
    corrSinSout = correlation_function(sin, sout, lags, mean_subtract=True, normalize=False)
    corrSinSin = correlation_function(sin, sin, lags, mean_subtract=True, normalize=False)
    corrSoutSout = correlation_function(sout, sout, lags, mean_subtract=True, normalize=False)
    win = hann(2*lag+1)

    
    if lag == 0:
        h = corrSinSout/corrSinSin
        fvals = 0
        gf = corrSinSout**2/(corrSinSin*corrSoutSout)
    else:
        # Normalize in the frequency domain
        corrSinSoutF = fft(corrSinSout*win)
        corrSinSinF = fft(corrSinSin*win)
        corrSoutSoutF = fft(corrSoutSout*win)
        hF = corrSinSoutF/np.abs(corrSinSinF)
        gf = np.abs(corrSinSoutF*corrSinSoutF.conj())/(np.abs(corrSinSinF)*np.abs(corrSoutSoutF))
        fvals = fftfreq(len(corrSinSout))
        h = ifft(hF)

# Plots for debugging/analyzing
    if debug:       
        # Time domain plots
        plt.figure()
        plt.subplot(141)
        plt.plot(lags, corrSinSout*win)
        plt.title('Cross-Corr')
        plt.subplot(142)
        plt.plot(lags, corrSinSin*win)
        plt.title('Auto-Corr Input')
        plt.subplot(143)
        plt.plot(lags, corrSoutSout*win)
        plt.title('Auto-Corr Output')
        plt.subplot(144)
        plt.plot(lags, h)
        plt.title('Filter')
    
        # Frequency domain plots
        plt.figure()
        fmid = len(fvals)//2
        plt.subplot(131)
        plt.plot(fvals[0:fmid], abs(corrSinSinF[0:fmid]) )
        plt.title('Input Power')
        plt.subplot(132)
        plt.plot(fvals[0:fmid], abs(corrSoutSoutF[0:fmid]) )
        plt.title('Output Power')
        plt.subplot(133)
        plt.plot(fvals[0:fmid], gf[0:fmid])
        plt.title('Coherence')
        
    return h, lags, gf, fvals
        

    
def coherency(s1, s2, lags, plot=False, window_fraction=None, noise_floor_db=None):
    """ Compute the coherency between two signals s1 and s2.

    :param s1: The first signal.
    :param s2: The second signal.
    :param lags: The lags to compute the coherency. They must be symmetric around zero, like lags=np.arange(-10, 11, 1).
    :param window_fraction: If not None, then each correlation function and auto-correlation-function is multiplied
            by a Gaussian window with standard deviation=window_fraction*lags.max(), prior to being turned into the
            coherency. This maybe suppresses high frequency noise in the coherency function.
    :param noise_floor_db: The threshold in decibels to zero out power in the auto and cross correlation function
            power spectrums, prior to taking the inverse FFT to produce the coherence. This is another way of
            eliminating high frequency noise in the coherency.

    :return: coh - The lags used to compute the coherency in units of time steps, and the coherency function.
    """

    # test for symmetry
    i = len(lags) // 2
    assert lags[i] == 0, "Midpoint of lags must be zero for coherency!"
    assert np.sum(-lags[:i] != lags[-i:][::-1]) == 0, "lags must be symmetric for coherency!"

    window = np.ones([len(lags)], dtype='float')
    if window_fraction is not None:
        assert window_fraction > 0 and window_fraction <= 1, "window_fraction must be between 0 and 1"
        # create a gaussian windowing function for the CF and ACFs
        window = np.exp(-lags**2 / (window_fraction*lags.max())**2)

    # do an FFT shift to the lags and the window, otherwise the FFT of the ACFs is not equal to the power
    # spectrum for some numerical reason
    window = fftshift(window)
    shift_lags = fftshift(lags)
    if len(lags) % 2 == 1:
        # shift zero from end of shift_lags to beginning
        shift_lags = np.roll(shift_lags, 1)
        window = np.roll(window, 1)

    cf = correlation_function(s1, s2, shift_lags)
    acf1 = correlation_function(s1, s1, shift_lags)
    acf2 = correlation_function(s2, s2, shift_lags)

    if np.sum(np.isnan(cf)) > 0:
        # print 'len(lags)=%d, len(s1)=%d, len(s2)=%d' % (len(lags), len(s1), len(s2))
        print('signals=',zip(s1, s2))
        print('shift_lags,cf=',zip(shift_lags, cf))
        raise Exception("Nans in cf")

    assert np.sum(np.isnan(acf1)) == 0, "Nans in acf1"
    assert np.sum(np.isnan(acf2)) == 0, "Nans in acf2"

    if window_fraction is not None:
        cf *= window
        acf1 *= window
        acf2 *= window

    cf_fft = fft(cf)
    acf1_fft = fft(acf1)
    acf2_fft = fft(acf2)

    acf1_ps = np.abs(acf1_fft)
    acf2_ps = np.abs(acf2_fft)

    # determine which points are noise (with magnitudes too low to be useful) in the acfs
    zeros = np.zeros([len(cf_fft)], dtype='bool')
    if noise_floor_db is not None:
        db1 = 20*np.log10(acf1_ps / acf1_ps.max()) + noise_floor_db
        z1 = db1 <= 0

        db2 = 20*np.log10(acf2_ps / acf2_ps.max()) + noise_floor_db
        z2 = db2 <= 0
        zeros = z1 | z2

    assert np.abs(acf1_fft.imag).max() < 1e-8, "acf1_fft.imag.max()=%f" % np.abs(acf1_fft.imag).max()
    assert np.abs(acf2_fft.imag).max() < 1e-8, "acf2_fft.imag.max()=%f" % np.abs(acf2_fft.imag).max()

    cpre = cf_fft / np.sqrt(acf1_ps*acf2_ps)
    cpre[zeros] = 0
    c = ifft(cpre)
    assert np.abs(c.imag).max() < 1e-8, "np.abs(c.imag).max()=%f" % np.abs(c.imag).max()

    coh = fftshift(c.real)
    freq = fftshift(fftfreq(len(lags)))
    fi = freq >= 0

    if np.sum(np.abs(coh) > 1) > 0:
        print('Warning: coherency is > 1!')

    if plot:
        plt.figure()
        plt.subplot(2, 3, 1)
        plt.plot(s1, 'r-')
        plt.plot(s2, 'b-')
        plt.legend(['s1', 's2'])
        plt.xlabel('Time')
        plt.axis('tight')
        plt.title('Signals')

        plt.subplot(2, 3, 2)
        plt.axvline(0, c='k')
        plt.axhline(0, c='k')
        l1 = plt.plot(lags, fftshift(acf1), 'r-')
        l2 = plt.plot(lags, fftshift(acf2), 'b-')
        l3 = plt.plot(lags, fftshift(cf), 'g-')
        plt.title('Correlation Functions')
        plt.xlabel('Lags')
        plt.legend(['', '', 'ACF1', 'ACF2', 'CF12'])
        plt.axis('tight')
        plt.ylim(-0.5, 1.0)

        plt.subplot(2, 3, 3)
        plt.axhline(0, c='k', alpha=0.75)
        plt.axvline(0, c='k', alpha=0.75)
        plt.plot(lags, coh, 'm-')
        plt.ylabel('Coherency')
        plt.xlabel('Lag')
        plt.axis('tight')
        plt.title('Coherency')

        plt.subplot(2, 3, 4)
        plt.plot(freq[fi], fftshift(acf1_ps)[fi], 'r')
        plt.plot(freq[fi], fftshift(acf2_ps)[fi], 'b')
        cf_ps = fftshift(np.abs(cf_fft))
        cf_pre_ps = fftshift(np.abs(cpre))
        plt.plot(freq[fi], cf_ps[fi], 'g--')
        plt.plot(freq[fi], cf_pre_ps[fi], 'm-')
        plt.legend(['ACF1', 'ACF2', 'CF12', 'CPRE'])
        plt.ylabel('Power (raw)')
        plt.xlabel('Frequency')
        plt.axis('tight')
        plt.title('Raw Power Spectra')

        if noise_floor_db:
            plt.subplot(2, 3, 5)
            plt.axhline(0, c='k')
            plt.plot(freq[fi], fftshift(db1)[fi], 'r')
            plt.plot(freq[fi], fftshift(db2)[fi], 'b')
            plt.legend(['ACF1', 'ACF2'])
            plt.ylabel('Power (dB)')
            plt.xlabel('Frequency')
            plt.axis('tight')
            plt.title('Log Power Spectra')

        plt.show()

    return coh


def get_envelope_end(env):
    """ Given an amplitude envelope, get the index that indicates the derivative of the envelope
        has converged to zero, indicating an end point.
    """
    denv = np.diff(env)
    i = np.where(np.abs(denv) > 0)[0]
    true_stop_index = np.max(i)+1
    return true_stop_index


def simple_smooth(s, window_len):

    w = np.hanning(window_len)
    w /= w.sum()
    return convolve1d(s, w)


def temporal_smooth(s, sample_rate, tau, hwinlen=20):
    """ Smooth with a gaussian.

    :param s: The signal
    :param tau: SD of gaussian
    :param hwinlen: Half the number of points used in the window.
    :return:  The smoothed signal
    """

    t = np.arange(-hwinlen, hwinlen+1) / sample_rate
    w = np.exp(-t**2 / tau)
    w /= w.sum()
    return convolve1d(s, w)


def quantify_cf(lags, cf, plot=False):
    """ Quantify properties of an auto or cross correlation function. """

    # identify the peak magnitude
    abs_cf = np.abs(cf)
    peak_magnitude = abs_cf.max()

    # identify the peak delay
    imax = abs_cf.argmax()
    peak_delay = lags[imax]

    # compute the area under the curve
    dt = np.diff(lags).max()
    cf_width = abs_cf.sum()*dt

    # compute the skewdness
    p = abs_cf / abs_cf.sum()
    mean = np.sum(lags*p)
    std = np.sqrt(np.sum(p*(abs_cf - mean)**2))
    skew = np.sum(p*(abs_cf - mean)**3) / std**3

    # compute the left and right areas under the absolute curve
    max_width = abs_cf[lags != 0].sum()*dt
    right_width = abs_cf[lags > 0].sum()*dt
    left_width = abs_cf[lags < 0].sum()*dt

    # create a measure of anisotropy from the AUCs
    anisotropy = (right_width - left_width) / max_width
    
    li = lags < 0
    ri = lags > 0

    # determine the mean lag time, i.e. the lag "center of mass". do this for each half
    cfl = np.abs(cf[li]) / np.abs(cf[li]).sum()
    left_lag = np.sum(cfl*lags[li])
    cfr = np.abs(cf[ri]) / np.abs(cf[ri]).sum()
    right_lag = np.sum(cfr*lags[ri])

    # integrate the right and left sides independently
    dl = np.diff(lags).max()
    left_sum = cf[li].sum()*dl
    right_sum = cf[ri].sum()*dl

    # take the correlation coefficient at zero lag
    cc = cf[lags == 0][0]

    if plot:
        plt.figure()
        plt.axhline(0, c='k')
        plt.plot(lags, cf, 'r-', linewidth=3)
        plt.axvline(peak_delay, c='g', alpha=0.75)
        plt.ylim(-1, 1)
        plt.axis('tight')
        t = 'width=%0.1f, mean=%0.1f, std=%0.1f, skew=%0.1f, anisotropy=%0.2f' % (cf_width, mean, std, skew, anisotropy)
        plt.title(t)
        plt.show()

    return {'magnitude':peak_magnitude, 'delay':peak_delay, 'width':cf_width,
            'mean':mean, 'std':std, 'skew':skew, 'anisotropy':anisotropy,
            'left_lag':left_lag, 'right_lag':right_lag, 'left_sum':left_sum, 'right_sum':right_sum, 'cc':cc}


def whiten(s, order):
    """ Whiten the signal s with an auto-regressive model of order specified by "order".

        :returns sw,coef sw is the whitened signal (original signal minus prediction), coef is coefficients of AR model
    """

    # remove the mean of the signal
    sm = s - s.mean()

    # construct a feature matrix
    X = np.zeros([len(s)-1, order])
    for k in range(order):
        X[k:, k] = sm[k:-1]

    # do a regression
    y = sm[1:]

    reg = Ridge(alpha=0, fit_intercept=False)
    reg.fit(X, y)
    spred = reg.predict(X)

    return sm - np.r_[0, spred], reg.coef_





from __future__ import division, print_function

import time

from abc import ABCMeta,abstractmethod
import numpy as np


from scipy.fftpack import fft,fftfreq, ifft
from scipy.signal import hilbert
from scipy.interpolate import RectBivariateSpline

import matplotlib.pyplot as plt

import nitime.algorithms as ntalg
from nitime import utils as ntutils
from soundsig.signal import lowpass_filter, bandpass_filter, correlation_function

# from brian import hears, Hz


class ComplexSpectrumEstimator(object):
    __metaclass__ = ABCMeta

    def __init__(self):
        pass

    @abstractmethod
    def estimate(self, signal, sample_rate, start_time, end_time):
        return NotImplementedError('Use a subclass of PowerSpectrumEstimator!')

    @abstractmethod
    def get_frequencies(self, signal_length, sample_rate):
        return NotImplementedError('Use a subclass of PowerSpectrumEstimator!')


class GaussianSpectrumEstimator(ComplexSpectrumEstimator):

    def __init__(self, nstd=6):
        ComplexSpectrumEstimator.__init__(self)
        self.nstd = nstd

    def get_frequencies(self, signal_length, sample_rate):
        freq = fftfreq(signal_length, d=1.0/sample_rate)
        nz = freq >= 0.0
        return freq[nz]

    def estimate(self, signal, sample_rate, start_time, end_time):
        nwinlen = len(signal)
        if nwinlen % 2 == 0:
            nwinlen += 1
        hnwinlen = nwinlen // 2

        #construct the window
        gauss_t = np.arange(-hnwinlen, hnwinlen+1, 1.0)
        gauss_std = float(nwinlen) / float(self.nstd)
        self.gauss_window = np.exp(-gauss_t**2 / (2.0*gauss_std**2)) / (gauss_std*np.sqrt(2*np.pi))

        #window the signal and take the FFT
        fft_len = len(signal)
        windowed_slice = signal[:fft_len]*self.gauss_window[:fft_len]
        s_fft = fft(windowed_slice, n=fft_len, overwrite_x=1)
        freq = fftfreq(fft_len, d=1.0/sample_rate)
        nz = freq >= 0.0

        return freq[nz],s_fft[nz]


class MultiTaperSpectrumEstimator(ComplexSpectrumEstimator):

    def __init__(self, bandwidth, adaptive=False, jackknife=False, max_adaptive_iter=150):
        ComplexSpectrumEstimator.__init__(self)
        self.bandwidth = bandwidth
        self.jackknife = jackknife
        self.adaptive = adaptive
        self.max_adaptive_iter = max_adaptive_iter

    def get_frequencies(self, signal_length, sample_rate):
        cspec_freq = fftfreq(signal_length, d=1.0/sample_rate)
        nz = cspec_freq >= 0.0
        return cspec_freq[nz]

    def estimate(self, signal, sample_rate, start_time, end_time, debug=False):

        slen = len(signal)

        #compute DPSS tapers for signals
        NW = max(1, int((slen / sample_rate)*self.bandwidth))
        K = 2*NW - 1

        tapers, eigs = ntalg.dpss_windows(slen, NW, K)
        ntapers = len(tapers)
        if debug:
            print('[MultiTaperSpectrumEstimator.estimate] slen=%d, NW=%d, K=%d, bandwidth=%0.1f, ntapers: %d' % (slen, NW, K, self.bandwidth, ntapers))

        #compute a set of tapered signals
        s_tap = tapers * signal

        #compute the FFT of each tapered signal
        s_fft = fft(s_tap, axis=1)

        #throw away negative frequencies of the spectrum
        cspec_freq = fftfreq(slen, d=1.0/sample_rate)
        nz = cspec_freq >= 0.0
        s_fft = s_fft[:, nz]
        flen = nz.sum()
        cspec_freq = cspec_freq[nz]
        #print '(1)cspec_freq.shape=',cspec_freq.shape
        #print '(1)s_fft.shape=',s_fft.shape

        #determine the weights used to combine the tapered signals
        if self.adaptive and ntapers > 1:
            #compute the adaptive weights
            weights,weights_dof = ntutils.adaptive_weights(s_fft, eigs, sides='twosided', max_iter=self.max_adaptive_iter)
        else:
            weights = np.ones([ntapers, flen]) / float(ntapers)

        #print '(1)weights.shape=',weights.shape

        def make_spectrum(signal, signal_weights):
            denom = (signal_weights**2).sum(axis=0)
            return (np.abs(signal * signal_weights)**2).sum(axis=0) / denom

        if self.jackknife:
            #do leave-one-out cross validation to estimate the complex mean and standard deviation of the spectrum
            cspec_mean = np.zeros([flen], dtype='complex')
            for k in range(ntapers):
                index = range(ntapers)
                del index[k]
                #compute an estimate of the spectrum using all but the kth weight
                cspec_est = make_spectrum(s_fft[index, :], weights[index, :])
                cspec_diff = cspec_est - cspec_mean
                #do an online update of the mean spectrum
                cspec_mean += cspec_diff / (k+1)
        else:
            #compute the average complex spectrum weighted across tapers
            cspec_mean = make_spectrum(s_fft, weights)

        return cspec_freq,cspec_mean.squeeze()


class WaveletSpectrumEstimator(ComplexSpectrumEstimator):

    def __init__(self, frequencies=None, num_cycles_per_window=10, min_freq=1, max_freq=512, num_freqs=20, nstd=6):
        ComplexSpectrumEstimator.__init__(self)

        self.num_cycles_per_window = num_cycles_per_window

        if frequencies is None:
            self.frequencies = np.logspace(np.log2(min_freq), np.log2(max_freq), num_freqs, base=2)
            self.frequencies = self.frequencies[::-1]
        else:
            self.frequencies = np.array(frequencies).astype('float')

        # determine window size for each center frequency
        self.window_lengths = np.zeros([len(self.frequencies)])
        self.standard_deviations = np.zeros([len(self.frequencies)])
        for k,f in enumerate(self.frequencies):
            # compute the standard deviation of a Gaussian that captures the right number of cycles for the frequency
            self.standard_deviations[k] = num_cycles_per_window / (nstd*f)
            self.window_lengths[k] = nstd*self.standard_deviations[k]

    def get_window_length(self):
        # return the largest window length for the timefreq function to use
        return self.window_lengths.max()

    def get_frequencies(self, signal_length, sample_rate):
        return self.frequencies

    def estimate(self, signal, sample_rate, start_time, end_time):

        # the maximum window length is used for each frequency, but the std of the gaussian changes
        slen = len(signal)

        # go through each frequency, window with a Gaussian and then multiply by complex exponential
        z = np.zeros([len(self.frequencies)], dtype='complex')
        for k,f in enumerate(self.frequencies):

            # construct the window
            t = np.linspace(start_time, end_time, slen)
            ct = ((end_time - start_time) / 2.0) + start_time
            gauss_window = np.exp(-(t - ct)**2 / self.standard_deviations[k]**2)

            # construct a complex exponential
            theta = 2*np.pi*f*(t - ct)
            cexp = np.cos(theta) + complex(0, 1)*np.sin(theta)

            # multiply the window, the complex exponential, and the signal, then sum them
            z[k] = np.sum(gauss_window*signal*cexp)*np.sqrt(f)

        return self.frequencies,z


def timefreq(s, sample_rate, window_length, increment, spectrum_estimator, min_freq=0, max_freq=None, zero_pad=True):
    """
        Compute a time-frequency representation of the signal s.

        s: the raw waveform.
        sample_rate: the sample rate of the waveform
        increment: the spacing in seconds between points where the spectrum is computed, i.e. inverse of the spectrogram sample rate
        spectrum_estimator: an instance of PowerSpectrumEstimator
        min_freq: the minimum frequency to analyze (Hz)
        max_freq: the maximum frequency to analyze (Hz)
        window_length: The length in seconds of the window to use for each segment. If None, then the spectrum_estimator
            object is queried.

        Returns t,freq,spec,rms:

        t: the time axis of the spectrogram
        freq: the frequency axis of the spectrogram
        tf: the time-frequency representation
    """

    if max_freq is None:
        max_freq = sample_rate / 2.0

    if window_length is None:
        window_length = spectrum_estimator.get_window_length()

    #compute lengths in # of samples
    nwinlen = int(sample_rate*window_length)
    if nwinlen % 2 == 0:
        nwinlen += 1
    hnwinlen = nwinlen // 2
    assert len(s) > nwinlen, "len(s)=%d, nwinlen=%d" % (len(s), nwinlen)

    # get the values for the frequency axis by estimating the spectrum of a dummy slice
    full_freq = spectrum_estimator.get_frequencies(nwinlen, sample_rate)
    freq_index = (full_freq >= min_freq) & (full_freq <= max_freq)
    freq = full_freq[freq_index]
    nfreq = freq_index.sum()

    nincrement = int(np.round(sample_rate*increment))
    if zero_pad:
        nwindows = len(s) // nincrement
        # print 'len(s)=%d, nwinlen=%d, hwinlen=%d, nincrement=%d, nwindows=%d' % (len(s), nwinlen, hnwinlen, nincrement, nwindows)
        #pad the signal with zeros
        zs = np.zeros([len(s) + 2*hnwinlen])
        zs[hnwinlen:-hnwinlen] = s
        window_centers = np.arange(nwindows)*nincrement + hnwinlen
    else:
        first_window_center = hnwinlen+1
        last_window_center = len(s)-hnwinlen
        window_centers = np.arange(first_window_center, last_window_center, nincrement, dtype='int')
        nwindows = len(window_centers)
        assert nwindows > 0, "nwindows=0, len(s)=%d, nwinlen=%d, nincrement=%d, window_centers=%s" % (len(s), nwinlen, nincrement, str(window_centers))
        zs = s
        assert window_centers.min() >= hnwinlen, "window_centers.minmax=(%d,%d), hnwinlen=%d, len(s)=%d" % (window_centers.min(), window_centers.max(), hnwinlen, len(s))
        assert window_centers.max() < len(s)-hnwinlen, "window_centers.minmax=(%d,%d), hnwinlen=%d, len(s)=%d" % (window_centers.min(), window_centers.max(), hnwinlen, len(s))

    #take the FFT of each segment, padding with zeros when necessary to keep window length the same
    #tf = np.zeros([nfreq, nwindows], dtype='complex')
    tf = np.zeros([nfreq, nwindows], dtype='complex')
    for k in range(nwindows):
        center = window_centers[k]
        si = center - hnwinlen
        ei = center + hnwinlen + 1

        spec_freq,est = spectrum_estimator.estimate(zs[si:ei], sample_rate, si/sample_rate, ei/sample_rate)
        findex = (spec_freq <= max_freq) & (spec_freq >= min_freq)
        #print 'k=%d' % k
        #print 'si=%d, ei=%d' % (si, ei)
        #print 'spec_freq.shape=',spec_freq.shape
        #print 'tf.shape=',tf.shape
        #print 'est.shape=',est.shape
        tf[:, k] = est[findex]

    # Note that the desired spectrogram rate could be slightly modified
    t = np.arange(0, nwindows, 1.0) * float(nincrement)/sample_rate

    return t, freq, tf


def generate_sliding_windows(N, sample_rate, increment, window_length):
    """
        Generate a list of indices representing windows into a signal of length N.
    """

    #compute lengths in # of samples
    nwinlen = int(sample_rate*window_length)
    if nwinlen % 2 == 0:
        nwinlen += 1
    hnwinlen = nwinlen // 2

    nincrement = int(np.round(sample_rate*increment))
    nwindows = N // nincrement

    windows = list()
    for k in range(nwindows):
        center = k*nincrement
        si = center - hnwinlen
        ei = center + hnwinlen + 1
        windows.append( (center, si, ei) )

    #compute the centers of each window
    t = np.arange(0, nwindows, 1.0) * increment

    return t, np.array(windows)


def gaussian_stft(s, sample_rate, window_length, increment, min_freq=0,
                  max_freq=None, nstd=6, zero_pad=True):
    """
        Compute a gaussian-windowed short-time fourier transform representation of the signal s.

        s: the raw waveform.
        sample_rate: the sample rate of the waveform
        window_length: The length in seconds of the window to use for each segment.
        increment: the spacing in seconds between points where the spectrum is computed, i.e. inverse of the spectrogram sample rate
        min_freq: the minimum frequency to analyze (Hz)
        max_freq: the maximum frequency to analyze (Hz) (if None then the nyquist frequency will be used)

        Returns t,freq,spec,rms:

        t: the time axis of the spectrogram
        freq: the frequency axis of the spectrogram
        tf: the time-frequency representation
    """

    spectrum_estimator = GaussianSpectrumEstimator(nstd=nstd)
    t, freq, tf = timefreq(s, sample_rate, window_length, increment,
                           spectrum_estimator=spectrum_estimator,
                           min_freq=min_freq, max_freq=max_freq,
                           zero_pad=zero_pad)
    ps = np.abs(tf)
    rms = ps.sum(axis=0)
    return t, freq, tf, rms


def mt_stft(s, sample_rate, window_length, increment, bandwidth=None, min_freq=0, max_freq=None, adaptive=True, jackknife=False):
    spectrum_estimator = MultiTaperSpectrumEstimator(bandwidth=bandwidth, adaptive=adaptive, jackknife=jackknife)
    return timefreq(s, sample_rate, window_length, increment, spectrum_estimator=spectrum_estimator, min_freq=min_freq, max_freq=max_freq)


def wavelet_scalogram(s, sample_rate, increment, frequencies=None, min_freq=1, max_freq=512, num_cycles_per_window=10,
                      num_freqs=20, nstd=6):
    """ Create a time frequency representation of the signal s by constructing a wavelet scalogram. Morelet wavelets
        are used.

    :param s: The signal, a 1D array.
    :param sample_rate: The sample rate of the signal in Hz.
    :param increment: The spacing in seconds between points where the windowed wavelet convolution is taken.
    :param frequencies: The center frequencies of each wavelet.
    :param min_freq: If frequencies is not specified, then the minimum center frequency to analyze.
    :param max_freq: If frequencies is not specified, then the maximum center frequency to analyze.
    :param num_cycles_per_window: The number of cycles for a given wavelet in a window.
    :param num_freqs: If frequencies is not specified, then the number of center frequencies between min_freq and max_freq.
    :param nstd: The number of standard deviations to consider for the Gaussian window when keeping the number of cycles constant.

    :return: (t, freq, tf) - t is a vector of time points, freq is an array of frequencies, tf is the time-frequency scalogram.
    """
    est = WaveletSpectrumEstimator(frequencies=frequencies, num_cycles_per_window=num_cycles_per_window,
                                   min_freq=min_freq, max_freq=max_freq, num_freqs=num_freqs, nstd=nstd)
    return timefreq(s, sample_rate, None, increment, spectrum_estimator=est, min_freq=min_freq, max_freq=max_freq)


class TimeFrequencyReassignment(object):

    def __init__(self):
        pass

    @abstractmethod
    def reassign(self, spec_t, spec_freq, spec):
        raise NotImplementedError('Use a subclass!')


class AmplitudeReassignment(object):
    """
        NOTE: doesn't work...
    """

    def __init__(self):
        pass

    def reassign(self, spec_t, spec_f, spec):

        #get power spectrum
        ps = np.abs(spec)

        #take the spectral and temporal derivatives
        dt = spec_t[1] - spec_t[0]
        df = spec_f[1] - spec_f[0]
        ps_df,ps_dt = np.gradient(ps)
        ps_df /= df
        ps_dt /= dt

        plt.figure()
        plt.subplot(2, 1, 1)
        plt.hist(ps_df.ravel(), bins=15)
        plt.title('ps_df')
        plt.axis('tight')
        plt.subplot(2, 1, 2)
        plt.hist(ps_dt.ravel(), bins=15)
        plt.title('ps_dt')
        plt.axis('tight')

        #construct the empty reassigned time frequency representation
        ps_r = np.zeros_like(ps)
        for k,freq in enumerate(spec_f):
            for j,t in enumerate(spec_t):
                inst_freq = ps_df[k, j]
                group_delay = ps_dt[k, j]
                print('inst_freq=%0.6f, group_delay=%0.6f' % (inst_freq, group_delay))
                fnew = freq + inst_freq
                tnew = group_delay + t
                print('fnew=%0.0f, tnew=%0.0f' % (fnew, tnew))
                row = np.array(np.nonzero(spec_f <= fnew)).max()
                col = np.array(np.nonzero(spec_t <= tnew)).max()
                print('row=',row)
                print('col=',col)
                ps_r[row, col] += 1.0

        ps_r /= len(spec_t)*len(spec_f)

        return ps_r


class PhaseReassignment(object):
    """
        NOTE: doesn't work...
    """

    def __init__(self):
        pass

    def reassign(self, spec_t, spec_f, spec):

        #get phase
        phase = np.angle(spec)

        #take the spectral and temporal derivatives
        dt = spec_t[1] - spec_t[0]
        df = spec_f[1] - spec_f[0]
        ps_df,ps_dt = np.gradient(phase)
        ps_df /= df
        ps_dt /= dt

        plt.figure()
        plt.subplot(2, 1, 1)
        plt.hist(ps_df.ravel(), bins=15)
        plt.title('ps_df')
        plt.axis('tight')
        plt.subplot(2, 1, 2)
        plt.hist(ps_dt.ravel(), bins=15)
        plt.title('ps_dt')
        plt.axis('tight')

        #construct the empty reassigned time frequency representation
        ps_r = np.zeros_like(phase)

        for k,freq in enumerate(spec_f):
            for j,t in enumerate(spec_t):
                tnew = max(0, t - (ps_df[k, j] / (2*np.pi)))
                fnew = max(0, ps_dt[k, j] / (2*np.pi))
                print('fnew=%0.0f, tnew=%0.0f' % (fnew, tnew))
                row = np.array(np.nonzero(spec_f <= fnew)).max()
                col = np.array(np.nonzero(spec_t <= tnew)).max()
                print('row=',row)
                print('col=',col)
                ps_r[row, col] += 1.0

        ps_r /= len(spec_t)*len(spec_f)

        return ps_r


def log_spectrogram(spec, offset=100):
    """
        Compute the log spectrogram.
    """
    lspec = np.zeros_like(spec)

    nz = spec > 0.0
    lspec[nz] = np.log10(spec[nz])
    lspec[nz] *= 10
    lspec[nz] += offset
    lspec[lspec < 0] = 0

    return lspec


def bandpass_timefreq(s, frequencies, sample_rate):
    """
        Bandpass filter signal s at the given frequency bands, and then use the Hilber transform
        to produce a complex-valued time-frequency representation of the bandpass filtered signal.
    """

    freqs = sorted(frequencies)
    tf_raw = np.zeros([len(frequencies), len(s)], dtype='float')
    tf_freqs = list()

    for k,f in enumerate(freqs):
        #bandpass filter signal
        if k == 0:
            tf_raw[k, :] = lowpass_filter(s, sample_rate, f)
            tf_freqs.append( (0.0, f) )
        else:
            tf_raw[k, :] = bandpass_filter(s, sample_rate,  freqs[k-1], f)
            tf_freqs.append( (freqs[k-1], f) )

    #compute analytic signal
    tf = hilbert(tf_raw, axis=1)
    #print 'tf_raw.shape=',tf_raw.shape
    #print 'tf.shape=',tf.shape

    return np.array(tf_freqs),tf_raw,tf


def gaussian_bandpass_analytic(s, sample_rate, frequencies, bandwidths, round=True):
    """ Compute the analytic signal in a set of bandpass channels
    :param s: the raw signal
    :param sample_rate: the signal's sample_rate
    :param frequencies: a list of center frequencies for gaussian filters
    :param bandwidths: a list of bandwidths for gaussian filters
    :param round: pad with zeros to next multiple of 2. Since we are doing so many iffts, this can speed things up a lot.

    :return analytic_signal: an array of dimension len(frequencies) * len(s)
    """

    if isinstance(bandwidths, (int, float)):
        bandwidths = [bandwidths]
    if len(bandwidths) != len(frequencies):
        if len(bandwidths) == 1:
            bandwidths = list(bandwidths) * len(frequencies)
        else:
            raise ValueError("bandwidths should be the same length as frequencies")

    # Enforce even window_length to have a symmetric window
    window_length = np.fix(np.fix(6 * sample_rate / (min(bandwidths) * 2.0 * np.pi)) / 2) * 2;

    if round:
        pow2_length = 2 ** np.ceil(np.log2(len(s) + window_length))
        window_length = (pow2_length - len(s)) / 2.0

    # Pad the input with zeros
    padded = np.pad(s,
                    (int(window_length / 2.0), int(np.ceil(window_length / 2.0))),
                    'constant')
    input_length = len(padded)
    input_start = int(window_length / 2.0)

    # Assign space for output
    analytic_signal = np.zeros((len(frequencies), len(s)), dtype=np.complex128)

    # Digital filtering
    spectrum = fft(padded)
    fft_freqs = fftfreq(input_length, d=(1.0 / sample_rate))
    nonzero_inds = fft_freqs >= 0
    positive_inds = fft_freqs > 0
    frequency_filter = np.zeros_like(spectrum)

    for ii, (freq, bw) in enumerate(zip(frequencies, bandwidths)):

        # Create the digital filter for this band
        frequency_filter[nonzero_inds] = np.exp(-0.5 * (fft_freqs[nonzero_inds] - freq) ** 2 / float(bw) ** 2)
        # Compute the filtered spectrum
        filtered_spectrum = frequency_filter * spectrum
        # Double the values of the positive frequencies for the analytic signal
        filtered_spectrum[positive_inds] *= 2

        # Compute the analytic signal
        bfinput = ifft(filtered_spectrum)
        # Remove the padding
        analytic_signal[ii] = bfinput[input_start: input_start + len(s)]

    return analytic_signal


def resample_spectrogram(t, freq, spec, dt_new, df_new):

    #print 'len(t)=%d, len(freq)=%d, spec.shape=(%d, %d)' % (len(t), len(freq), spec.shape[0], spec.shape[1])
    spline = RectBivariateSpline(freq, t, spec)

    ntnew = int(np.ceil((t.max() - t.min()) / dt_new))
    nfnew = int(np.ceil((freq.max() - freq.min()) / df_new))

    tnew = np.arange(ntnew)*dt_new
    fnew = np.arange(nfnew)*df_new

    new_spec = spline(fnew, tnew)

    return tnew,fnew,new_spec


def compute_mean_spectrogram(s, sample_rate, win_sizes, increment=None, num_freq_bands=100,
                             spec_estimator=GaussianSpectrumEstimator(nstd=6), mask=False, mask_gain=3.0):
    """
        Compute a spectrogram for each time window, and average across time windows to get better time-frequency
        resolution. Post-processing is done with applying the log to change the power spectrum to decibels, and
        then a hard threshold is applied to zero-out the lowest 10% of the pixels.
    """

    #compute spectrograms
    stime = time.time()
    timefreqs = list()
    for k,win_size in enumerate(win_sizes):
        if increment is None:
            inc = win_sizes[0] / 2  # TODO (kevin): i cant tell if this is supposed to floating point or integer division
        else:
            inc = increment
        t,freq,tf = timefreq(s, sample_rate, win_size, inc, spec_estimator)
        ps = np.abs(tf)
        ps_log = log_spectrogram(ps)
        timefreqs.append( (t, freq, ps_log) )
    etime = time.time() - stime
    #print 'time to compute %d spectrograms: %0.6fs' % (len(win_sizes), etime)

    #compute the mean spectrogram across window sizes
    nyquist_freq = sample_rate / 2.0
    df = nyquist_freq / num_freq_bands
    f_smallest = np.arange(num_freq_bands)*df
    t_smallest = timefreqs[0][0]  # best temporal resolution
    df_smallest = f_smallest[1] - f_smallest[0]
    dt_smallest = t_smallest[1] - t_smallest[0]

    #resample the spectrograms so they all have the same frequency spacing
    stime = time.time()
    rs_specs = list()
    for t,freq,ps in timefreqs:
        rs_t,rs_freq,rs_ps = resample_spectrogram(t, freq, ps, dt_smallest, df_smallest)
        rs_specs.append(rs_ps)
    etime = time.time() - stime
    #print 'time to resample %d spectrograms: %0.6fs' % (len(win_sizes), etime)

    #get the shortest spectrogram length
    min_freq_len = np.min([rs_ps.shape[0] for rs_ps in rs_specs])
    min_t_len = np.min([rs_ps.shape[1] for rs_ps in rs_specs])
    rs_specs_arr = np.array([rs_ps[:min_freq_len, :min_t_len] for rs_ps in rs_specs])
    t_smallest = np.arange(min_t_len)*dt_smallest
    f_smallest = np.arange(min_freq_len)*df_smallest

    #compute mean, std, and zscored power spectrum across window sizes
    tf_mean = rs_specs_arr.mean(axis=0)

    if mask:
        #compute the standard deviation across window sizes
        tf_std = rs_specs_arr.std(axis=0, ddof=1)
        #compute something that is close to the maximum std. we use the 95th pecentile to avoid outliers
        tf_std /= np.percentile(tf_std.ravel(), 95)
        #compute a sigmoidal mask that will zero out pixels in tf_mean that have high standard deviations
        sigmoid = lambda x: 1.0 / (1.0 + np.exp(1)**(-mask_gain*x))
        sigmoid_mask = 1.0 - sigmoid(tf_std)
        #mask the mean time frequency representation
        tf_mean *= sigmoid_mask

    return t_smallest, f_smallest, tf_mean


def define_f_bands(stt=180, stp=7000, n_bands=32, kind='log'):
    '''
    Defines log-spaced frequency bands...generally this is for auditory
    spectrogram extraction. Brian used 180 - 7000 Hz, so for now those
    are the defaults.

    INPUTS
    --------
        stt : int
            The starting frequency
        stp : int
            The end frequency
        n_bands : int
            The number of bands to calculate
        kind : string, ['log', 'erb']
            What kind of spacing will we use for the frequency bands.
    '''
    if kind == 'log':
        aud_fs = np.logspace(np.log10(stt), np.log10(stp), n_bands).astype(int)
    elif kind == 'lin':
        aud_fs = np.linspace(stt, stp, n_bands).astype(int)
 #   elif kind == 'erb':
 #       aud_fs = hears.erbspace(stt*Hz, stp*Hz, n_bands)
    else:
        raise NameError("I don't know what kind of spacing that is")
    return aud_fs


#def extract_nsl_spectrogram(sig, Fs, cfs):
#    '''Implements a version of the "wav2aud" function in the NSL toolbox.
#    Uses Brian hears to chain most of the computations to be done online.
#
#    This is effectively what it does:
#        1. Gammatone filterbank at provided cfs (erbspace recommended)
#        2. Half-wave rectification
#        3. Low-pass filtering at 2Khz
#        4. First-order derivative across frequencies (basically just
#            taking the diff of successive frequencies to sharpen output)
#        5. Half-wave rectification #2
#        6. An exponentially-decaying average, with time constant chosen
#            to be similar to that reported in the NSL toolbox (8ms)
#
#    INPUTS
#    --------
#    sig : array
#        The auditory signals we'll use to extract. Should be time x feats, or 1-d
#    Fs : float, int
#        The sampling rate of the signal
#    cfs : list of floats, ints
#        The center frequencies that we'll use for initial filtering.
#
#    OUTPUTS
#    --------
#    out : array, [tpts, len(cfs)]
#        The auditory spectrogram of the signal
#    '''
#    Fs = float(Fs)*Hz
#    snd = hears.Sound(sig, samplerate=Fs)
#
#    # Cochlear model
#    snd_filt = hears.Gammatone(snd, cfs)
#
#    # Hair cell stages
#    clp = lambda x: np.clip(x, 0, np.inf)
#    snd_hwr = hears.FunctionFilterbank(snd_filt, clp)
#    snd_lpf = hears.LowPass(snd_hwr, 2000)
#
#    # Lateral inhibitory network
#    rands = lambda x: sigp.roll_and_subtract(x, hwr=True)
#    snd_lin = hears.FunctionFilterbank(snd_lpf, rands)
#
#    # Initial processing
#    out = snd_lin.process()
#
#    # Time integration.
#    # Time constant is 8ms, which we approximate with halfwidth of 12
#    half_pt = (12. / 1000) * Fs
#    out = pd.stats.moments.ewma(out, halflife=half_pt)
#    return out


def roll_and_subtract(sig, amt=1, axis=1, hwr=False):
    '''Rolls the input matrix along the specifies axis, then
    subtracts this from the original signal. This is meant to
    be similar to the lateral inhibitory network from Shamma's
    NSL toolbox. hwr specifies whether to include a half-wave
    rectification after doing the subtraction.'''
    diff = np.roll(sig, -amt, axis=axis)
    diff[:, -amt:] = 0
    diff = np.subtract(sig, diff)
    if hwr is True:
        diff = np.clip(diff, 0, np.inf)
    return diff


def power_spectrum_jn(s, sample_rate, window_length, increment, min_freq=0, max_freq=None):
    """ Computes the power spectrum of a signal by averaging across time-frequency representation
        created using a Gaussian-windowed Short-time Fourier Transform. Uses jacknifing to estimate
        the variance of the spectra.

    :param s: The first signal
    :param sample_rate: The sample rates of the signal.
    :param window_length: The length of the window used to compute the STFT (units=seconds)
    :param increment: The spacing between the points of the STFT  (units=seconds)
    :param min_freq: The minimum frequency to analyze (units=Hz, default=0)
    :param max_freq: The maximum frequency to analyze (units=Hz, default=nyquist frequency)

    :return: freq,psd,psd_var,phase: freq is an array of frequencies that the spectrum was computed
             at. psd is an array of length len(freq) that contains the power at each frequency.
             psd_var is the variance of the power spectrum. phase is the phase at each frequency.
    """

    t, freq, tf, rms = gaussian_stft(s, sample_rate, window_length=window_length, increment=increment,
                                     min_freq=min_freq, max_freq=max_freq, zero_pad=False)
    ps = np.abs(tf)**2

    # make leave-one-out estimates of the spectrum
    jn_estimates = list()
    njn = tf.shape[1]
    for k in range(njn):
        i = np.ones([njn], dtype='bool')
        i[k] = False
        ps_k = tf[:, i].mean(axis=1)
        jn_estimates.append(ps_k)
    jn_estimates = np.array(jn_estimates)

    # estimate the variance of the coherence
    jn_mean = jn_estimates.mean(axis=0)
    jn_diff = (jn_estimates - jn_mean)**2
    ps_var = ((njn-1) / float(njn)) * jn_diff.sum(axis=0)

    # compute the spectrum using all the data
    ps_mean = ps.mean(axis=1)

    # compute the phase
    z = tf.sum(axis=1)
    phase = np.angle(z)

    return freq,ps_mean,ps_var,phase


def power_spectrum_from_acf(s, sample_rate, lags):
    """ Compute the power spectrum of a signal s by taking the FFT of the auto-correlation function.

    :param s: The signal.
    :param sample_rate: The sample rate of the signal s.
    :param lags: integer-valued lags, should be symmetric around zero.
    :return: freq,psd: The frequency of the power spectrum and the power spectrum
    """

    acf = correlation_function(s, s, lags, mean_subtract=True, normalize=True)

    psd = np.abs(fft(acf))**2
    freq = fftfreq(len(acf), d=1. / sample_rate)
    i = freq >= 0
    return freq[i], psd[i]




from __future__ import division, print_function

import time

import numpy as np
from numpy.fft import fftshift, ifft
from scipy import fftpack
import nitime.algorithms as ntalg
from nitime import utils as ntutils
from scipy.ndimage import convolve1d

from soundsig.signal import gaussian_window
from soundsig.spikes import compute_psth
from soundsig.timefreq import gaussian_stft


class CoherenceData(object):
    """
        This class encapsulates a bunch of data for coherence calculations:

        .frequency: an array of frequencies that the coherence is computed for
        .coherence: the coherence computed, same dimensionality as frequency
        .coherence_upper: the upper bound of the coherence
        .coherence_lower: the lower bound of the coherence
        .sample_rate: the sample rate in Hz of the original signal
        .frequency_cutoff: the frequency at which the lower bound of the coherence goes below zero
        .nmi: the normal mutual information in bits/s
    """

    def __init__(self, frequency_cutoff=None):
        self.frequency = None
        self.coherence = None
        self.coherence_upper = None
        self.coherence_lower = None
        self.sample_rate = None

        self.frequency_cutoff = frequency_cutoff
        self._nmi = None

    @property
    def nmi(self):
        if self._nmi is None and self.coherence is not None and self.frequency is not None and self.coherence_lower is not None:
            nfreq_cutoff,nnminfo = compute_freq_cutoff_and_nmi(self.frequency, self.sample_rate, self.coherence, self.coherence_lower, freq_cutoff=self.frequency_cutoff)
            self.frequency_cutoff = nfreq_cutoff
            self._nmi = nnminfo
        return self._nmi


def merge_spike_trials(trials_by_stim, stim_durations):
    """ Helper function - if you want to analyze  coherence across multiple stimuli, with
        multiple trials per stimulus, construct a list of lists called trials_by_stim, where:

            trials_by_stim[i] = [ [t1, t2, t3, t4, ...],
                                  [t1, t2, ..],
                                  [t1, t2, t3, t4, t5, ...],
                                  ...
                                ]
            i.e. each element of trials_by_stim is comprised a list of trials, each trial is a list of spike times.

            stim_durations is a list of stimulus durations in seconds, where len(stim_durations) == len(trials_by_stim)
    """

    assert len(trials_by_stim) == len(stim_durations)
    ntrials_all = np.array([len(spike_trials) for spike_trials in trials_by_stim])
    ntrials = ntrials_all.max()

    merged_spike_trials = list()
    for k in range(ntrials):
        merged_spike_trials.append(list())

    for k,spike_trials in enumerate(trials_by_stim):

        sd = 0.0
        if k > 0:
            sd = stim_durations[k-1]

        for m,spike_train in enumerate(spike_trials):
            merged_spike_trials[m].extend(np.array(spike_train) + sd)

    return merged_spike_trials


class ModelCoherenceAnalyzer(object):
    """
        You can use this object to compute the coherence upper bound for a set of spike trains,
        along with predictions for the PSTHs.
    """

    def __init__(self, spike_trials_by_stim, preds_by_stim, window_size=3.0, bandwidth=15.0, bin_size=0.001,
                 frequency_cutoff=None, tanh_transform=False):
        """
            spike_trials_by_stim: an array of arrays of arrays, the length is equal to the number of
                stimuli, and element i looks like:
                spike_trials_by_stim[i] = [ [t1, t2, t3, t4, ...],
                                            [t1, t2, ..],
                                            [t1, t2, t3, t4, t5, ...],
                                            ...
                                          ]

            preds_by_stim: a predicted PSTH for each stimulus, so len(preds_by_stim)=len(spike_trials_by_stim).

            window_size: the size in seconds of the window used to compute Fourier Transforms. The signal is
                broken up into segments of length window_size prior to computing the multi-tapered spectral density.

            bandwidth: a quantity related to the number of tapers used in the spectral density estimates

            bin_size: the size in seconds of the bins used to compute the PSTH

            frequency_cutoff: frequency in Hz to stop for computing the normal mutual information

            tanh_transform: whether to transform the coherence prior to computing the jacknifed upper bounds
        """

        if len(spike_trials_by_stim) != len(preds_by_stim):
            print('# of stims for spike trials should equal # of predictions')
            return

        psth_lens = np.array([len(pred) for pred in preds_by_stim])

        self.spike_trials_by_stim = spike_trials_by_stim
        self.window_size = window_size
        self.bin_size = bin_size
        self.bandwidth = bandwidth
        self.cdata_bound,self.ncdata_bound = compute_single_spike_coherence_bound(self.spike_trails_by_stim, self.bandwidth,
                                                                                  self.window_size, psth_lens, bin_size=self.bin_size,
                                                                                  frequency_cutoff=frequency_cutoff,
                                                                                  tanh_transform=tanh_transform)

        concat_preds = np.zeros([psth_lens.sum()])
        offset = 0
        for k,pred in enumerate(preds_by_stim):
            e = offset + psth_lens[k]
            concat_preds[offset:e] = pred
            offset = e

        self.model_data = compute_coherence_model_performance(self.spike_trails_by_stim, concat_preds, self.bandwidth,
                                                              self.window_size, psth_lens, bin_size=self.bin_size,
                                                              cdata_bound=self.cdata_bound, frequency_cutoff=frequency_cutoff,
                                                              tanh_transform=tanh_transform)


def get_concat_split_psths(spike_trials_by_stim, psth_lens, bin_size):
    """
        Takes an array of arrays of spike times, splits each into even and
        odd trials, and computes the PSTH for the even and odd trials.
    """

    N = psth_lens.sum()

    concat_even_psths = np.zeros([N])
    concat_odd_psths = np.zeros([N])

    offset = 0
    for m,spike_trials in enumerate(spike_trials_by_stim):
        even_trials = [ti for k,ti in enumerate(spike_trials) if k % 2]
        odd_trials = [ti for k,ti in enumerate(spike_trials) if not k % 2]
        duration = psth_lens[m] * bin_size
        even_psth = compute_psth(even_trials, duration, bin_size=bin_size)
        odd_psth = compute_psth(odd_trials, duration, bin_size=bin_size)

        e = offset + psth_lens[m]
        concat_even_psths[offset:e] = even_psth
        concat_odd_psths[offset:e] = odd_psth
        offset = e
    return concat_even_psths,concat_odd_psths


def get_concat_psth(spike_trials_by_stim, psth_lens, bin_size):
    """
        Takes a bunch of spike trials, separated by stimulus, creates a PSTH per stimulus,
        and concatenates each PSTH into a long array.
    """

    N = np.sum(psth_lens)

    concat_psths = np.zeros([N])

    offset = 0
    for k,spike_trials in enumerate(spike_trials_by_stim):
        duration = psth_lens[k] * bin_size
        psth = compute_psth(spike_trials, duration, bin_size=bin_size)
        e = offset + psth_lens[k]
        concat_psths[offset:e] = psth
        offset = e
    return concat_psths


def compute_single_spike_coherence_bound(spike_trials_by_stim, bandwidth, window_size, psth_lens, bin_size=0.001,
                                         frequency_cutoff=None, tanh_transform=False):
    """
        Computes the coherence between a set of spike trains and themselves. Useful for producing
        an upper bound on possible coherence in a model-independent way.

        spike_trials_by_stim: an array of arrays of arrays, the length is equal to the number of
                stimuli, and element i looks like:
                spike_trials_by_stim[i] = [ [t1, t2, t3, t4, ...],
                                            [t1, t2, ..],
                                            [t1, t2, t3, t4, t5, ...],
                                            ...
                                          ]

        preds_by_stim: a predicted PSTH for each stimulus, so len(preds_by_stim)=len(spike_trials_by_stim).

        psth_lens: The length in seconds of each stimulus. len(psth_lens) = len(spike_trials_by_stim)

        window_size: the size in seconds of the window used to compute Fourier Transforms. The signal is
            broken up into segments of length window_size prior to computing the multi-tapered spectral density.

        bandwidth: a quantity related to the number of tapers used in the spectral density estimates

        bin_size: the size in seconds of the bins used to compute the PSTH

        frequency_cutoff: frequency in Hz to stop for computing the normal mutual information

        tanh_transform: whether to transform the coherence prior to computing the jacknifed upper bounds

    """

    all_ntrials = np.array([len(spike_trials) for spike_trials in spike_trials_by_stim])
    ntrials = all_ntrials.max()
    even_psth,odd_psth = get_concat_split_psths(spike_trials_by_stim, psth_lens, bin_size)

    def cnormalize(c, num_trials):
        sign = np.sign(c)
        cabs = np.abs(c)
        index = cabs > 0.0
        kdown = (-num_trials + num_trials * np.sqrt(1.0 / cabs[index])) / 2.0
        kdown *= sign[index]
        cnorm = np.zeros(c.shape)
        cnorm[index] = 1.0 / (kdown + 1.0)
        return cnorm

    sample_rate = 1.0 / bin_size
    cdata = compute_mtcoherence(even_psth, odd_psth, sample_rate, window_size, bandwidth=bandwidth,
                              frequency_cutoff=frequency_cutoff, tanh_transform=tanh_transform)

    ncoherence_mean = cnormalize(cdata.coherence, ntrials)
    ncoherence_upper = cnormalize(cdata.coherence_upper, ntrials)
    ncoherence_lower = cnormalize(cdata.coherence_lower, ntrials)

    ncdata = CoherenceData(frequency_cutoff=frequency_cutoff)
    ncdata.frequency = cdata.frequency
    ncdata.sample_rate = sample_rate
    ncdata.coherence = ncoherence_mean
    ncdata.coherence_lower = ncoherence_lower
    ncdata.coherence_upper = ncoherence_upper

    return cdata,ncdata


def compute_coherence_model_performance(spike_trials_by_stim, psth_prediction, bandwidth, window_size, psth_lens, bin_size=0.001,
                                        cdata_bound=None, frequency_cutoff=None, tanh_transform=False):
    """
        Computes coherence of the spike trains themselves, but also the model. Use the ModelCoherenceAnalyzer class
        instead of calling this function directly.
    """

    all_ntrials = np.array([len(spike_trials) for spike_trials in spike_trials_by_stim])
    ntrials = all_ntrials.max()

    #compute upper bound for model performance from real data
    if cdata_bound is None:
        cdata_bound,ncdata_bound = compute_single_spike_coherence_bound(spike_trials_by_stim, bandwidth, window_size,
                                                                        psth_lens, bin_size=bin_size, tanh_transform=tanh_transform)

    psth = get_concat_psth(spike_trials_by_stim, psth_lens, bin_size=bin_size)

    sample_rate = 1.0 / bin_size

    #compute the non-normalized coherence between the real PSTH and the model prediction of the PSTH
    cdata_model = compute_mtcoherence(psth, psth_prediction, sample_rate, window_size, bandwidth=bandwidth, frequency_cutoff=frequency_cutoff)

    def cnormalize(cbound, cpred, num_trials):
        sign = np.sign(cbound)
        cbound_abs = np.abs(cbound)
        index = np.abs(cbound) > 0.0
        rhs = (1.0 + np.sqrt(1.0 / cbound_abs[index])) / (-num_trials + num_trials * np.sqrt(1.0 / cbound_abs[index]) + 2.0)
        rhs *= sign[index]
        ncpred = np.zeros(cpred.shape)
        ncpred[index] = cpred[index] * rhs
        return ncpred

    #correct each model coherence
    ncmean_model = cnormalize(cdata_bound.coherence, cdata_model.coherence, ntrials)
    ncupper_model = cnormalize(cdata_bound.coherence_upper, cdata_model.coherence_upper, ntrials)
    nclower_model = cnormalize(cdata_bound.coherence_lower, cdata_model.coherence_lower, ntrials)

    mcdata = CoherenceData(frequency_cutoff=frequency_cutoff)
    mcdata.frequency = cdata_model.frequency
    mcdata.sample_rate = sample_rate
    mcdata.coherence = ncmean_model
    mcdata.coherence_lower = nclower_model
    mcdata.coherence_upper = ncupper_model

    return mcdata


def compute_coherence_original(s1, s2, sample_rate, bandwidth, jackknife=False, tanh_transform=False):
    """
        An implementation of computing the coherence. Don't use this.
        TODO (kevin): What is this???
    """

    minlen = min(len(s1), len(s2))
    if s1.shape != s2.shape:
        s1 = s1[:minlen]
        s2 = s2[:minlen]

    window_length = len(s1) / sample_rate  # TODO (kevin): should this be integer or float?
    window_length_bins = int(window_length * sample_rate)

    #compute DPSS tapers for signals
    NW = int(window_length*bandwidth)
    K = 2*NW - 1
    print('compute_coherence: NW=%d, K=%d' % (NW, K))
    tapers,eigs = ntalg.dpss_windows(window_length_bins, NW, K)

    njn = len(eigs)
    jn_indices = [range(njn)]
    #compute jackknife indices
    if jackknife:
        jn_indices = list()
        for i in range(len(eigs)):
            jn = range(len(eigs))
            jn.remove(i)
            jn_indices.append(jn)

    #taper the signals
    s1_tap = tapers * s1
    s2_tap = tapers * s2

    #compute fft of tapered signals
    s1_fft = fftpack.fft(s1_tap, axis=1)
    s2_fft = fftpack.fft(s2_tap, axis=1)

    #compute adaptive weights for each taper
    w1,nu1 = ntutils.adaptive_weights(s1_fft, eigs, sides='onesided')
    w2,nu2 = ntutils.adaptive_weights(s2_fft, eigs, sides='onesided')

    coherence_estimates = list()
    for jn in jn_indices:

        #compute cross spectral density
        sxy = ntalg.mtm_cross_spectrum(s1_fft[jn, :], s2_fft[jn, :], (w1[jn], w2[jn]), sides='onesided')

        #compute individual power spectrums
        sxx = ntalg.mtm_cross_spectrum(s1_fft[jn, :], s1_fft[jn, :], w1[jn], sides='onesided')
        syy = ntalg.mtm_cross_spectrum(s2_fft[jn, :], s2_fft[jn, :], w2[jn], sides='onesided')

        #compute coherence
        coherence = np.abs(sxy)**2 / (sxx * syy)
        coherence_estimates.append(coherence)

    #compute variance
    coherence_estimates = np.array(coherence_estimates)
    coherence_variance = np.zeros([coherence_estimates.shape[1]])
    coherence_mean = coherence_estimates[0]
    if jackknife:
        coherence_mean = coherence_estimates.mean(axis=0)
        #mean subtract and square
        cv = np.sum((coherence_estimates - coherence_mean)**2, axis=0)
        coherence_variance[:] = (1.0 - 1.0/njn) * cv

    #compute frequencies
    sampint = 1.0 / sample_rate
    L = minlen // 2 + 1
    freq = np.linspace(0, 1.0 / (2.0 * sampint), L)

    #compute upper and lower bounds
    cmean = coherence_mean
    coherence_lower = cmean - 2*np.sqrt(coherence_variance)
    coherence_upper = cmean + 2*np.sqrt(coherence_variance)

    cdata = CoherenceData()
    cdata.coherence = coherence_mean
    cdata.coherence_lower = coherence_lower
    cdata.coherence_upper = coherence_upper
    cdata.frequency = freq
    cdata.sample_rate = sample_rate

    return cdata


def compute_mtcoherence(s1, s2, sample_rate, window_size, bandwidth=15.0, chunk_len_percentage_tolerance=0.30,
                      frequency_cutoff=None, tanh_transform=False, debug=False):
    """
        Computing the multi-taper coherence between signals s1 and s2. To do so, the signals are broken up into segments of length
        specified by window_size. Then the multi-taper coherence is computed between each segment. The mean coherence
        is computed across segments, and an estimate of the coherence variance is computed across segments.

        sample_rate: the sample rate in Hz of s1 and s2

        window_size: size of the segments in seconds

        bandwidth: related to the # of tapers used to compute the spectral density. The higher the bandwidth, the more tapers.

        chunk_len_percentage_tolerance: If there are leftover segments whose lengths are less than window_size, use them
            if they comprise at least the fraction of window_size specified by chunk_len_percentage_tolerance

        frequency_cutoff: the frequency at which to cut off the coherence when computing the normal mutual information

        tanh_transform: whether to transform the coherences when computing the upper and lower bounds, supposedly
            improves the estimate of variance.
    """

    minlen = min(len(s1), len(s2))
    if s1.shape != s2.shape:
        s1 = s1[:minlen]
        s2 = s2[:minlen]

    sample_length_bins = min(len(s1), int(window_size * sample_rate))

    #compute DPSS tapers for signals
    NW = int(window_size*bandwidth)
    K = 2*NW - 1
    #print('compute_coherence: NW=%d, K=%d' % (NW, K))
    tapers,eigs = ntalg.dpss_windows(sample_length_bins, NW, K)
    if debug:
        print('[compute_coherence] bandwidth=%0.1f, # of tapers: %d' % (bandwidth, len(eigs)))
        print(eigs)

    #break signal into chunks and estimate coherence for each chunk
    nchunks = int(np.floor(len(s1) / float(sample_length_bins)))
    nleft = len(s1) % sample_length_bins
    if nleft > 0:
        nchunks += 1
    #print('sample_length_bins=%d, # of chunks:%d, # samples in last chunk: %d' % (sample_length_bins, nchunks, nleft))
    coherence_estimates = list()
    for k in range(nchunks):
        s = k*sample_length_bins
        e = min(len(s1), s + sample_length_bins)
        chunk_len = e - s
        chunk_percentage = chunk_len / float(sample_length_bins)
        if chunk_percentage < chunk_len_percentage_tolerance:
            #don't compute coherence for a chunk whose length is less than a certain percentage of sample_length_bins
            continue
        s1_chunk = np.zeros([sample_length_bins])
        s2_chunk = np.zeros([sample_length_bins])
        s1_chunk[:chunk_len] = s1[s:e]
        s2_chunk[:chunk_len] = s2[s:e]

        #taper the signals
        s1_tap = tapers * s1_chunk
        s2_tap = tapers * s2_chunk

        #compute fft of tapered signals
        s1_fft = fftpack.fft(s1_tap, axis=1)
        s2_fft = fftpack.fft(s2_tap, axis=1)

        #compute adaptive weights for each taper
        w1,nu1 = ntutils.adaptive_weights(s1_fft, eigs, sides='onesided')
        w2,nu2 = ntutils.adaptive_weights(s2_fft, eigs, sides='onesided')

        #compute cross spectral density
        sxy = ntalg.mtm_cross_spectrum(s1_fft, s2_fft, (w1, w2), sides='onesided')

        #compute individual power spectrums
        sxx = ntalg.mtm_cross_spectrum(s1_fft, s1_fft, w1, sides='onesided')
        syy = ntalg.mtm_cross_spectrum(s2_fft, s2_fft, w2, sides='onesided')

        #compute coherence
        coherence = np.abs(sxy)**2 / (np.abs(sxx) * np.abs(syy))
        coherence_estimates.append(coherence)

    #compute variance
    coherence_estimates = np.array(coherence_estimates)

    if tanh_transform:
        coherence_estimates = np.arctanh(coherence_estimates)

    coherence_variance = np.zeros([coherence_estimates.shape[1]])
    coherence_mean = coherence_estimates.mean(axis=0)
    #mean subtract and square
    cv = np.sum((coherence_estimates - coherence_mean)**2, axis=0)
    coherence_variance[:] = (1.0 - 1.0/nchunks) * cv

    if tanh_transform:
        coherence_variance = np.tanh(coherence_variance)
        coherence_mean = np.tanh(coherence_mean)

    #compute frequencies
    sampint = 1.0 / sample_rate
    L = sample_length_bins // 2 + 1
    freq = np.linspace(0, 1.0 / (2.0 * sampint), L)

    #compute upper and lower bounds
    coherence_lower = coherence_mean - 2*np.sqrt(coherence_variance)
    coherence_upper = coherence_mean + 2*np.sqrt(coherence_variance)

    cdata = CoherenceData(frequency_cutoff=frequency_cutoff)
    cdata.coherence = coherence_mean
    cdata.coherence_lower = coherence_lower
    cdata.coherence_upper = coherence_upper
    cdata.frequency = freq
    cdata.sample_rate = sample_rate

    return cdata


def compute_freq_cutoff_and_nmi(freq, sample_rate, coherence_mean, coherence_lower, freq_cutoff=None):
    """
        Given the coherence and lower bound on coherence, compute the frequency cutoff, which is
        the point at which the lower bound dips below zero.
    """

    if freq_cutoff is None:
        #find frequency at which lower bound dips below zero
        zindices = np.where(coherence_lower <= 0.0)[0]
        freq_cutoff_index = len(coherence_mean)
        if len(zindices) > 0:
            freq_cutoff_index = min(zindices)
        else:
            zindices = np.where(freq < (sample_rate / 2.0))[0]
            if len(zindices) > 0:
                freq_cutoff_index = max(zindices)
        freq_cutoff = freq[freq_cutoff_index]
    else:
        freq_cutoff_index = max(np.where(freq <= freq_cutoff)[0])
        print(freq_cutoff_index)

    #compute normalized mutual information
    df = freq[1] - freq[0]
    nminfo = -df * np.log2(1.0 - coherence_mean[:freq_cutoff_index]).sum()

    return freq_cutoff,nminfo


def coherence_jn(s1, s2, sample_rate, window_length, increment, min_freq=0, max_freq=None, return_coherency=False):
    """ Computes the coherence between two signals by averaging across time-frequency representations
        created using a Gaussian-windowed Short-time Fourier Transform. Uses jacknifing to estimate
        the variance of the coherence.

    :param s1: The first signal
    :param s2: The second signal
    :param sample_rate: The sample rates of the signals.
    :param window_length: The length of the window used to compute the STFT (units=seconds)
    :param increment: The spacing between the points of the STFT  (units=seconds)
    :param min_freq: The minimum frequency to analyze (units=Hz, default=0)
    :param max_freq: The maximum frequency to analyze (units=Hz, default=nyquist frequency)
    :param return_coherency: Whether or not to return the time domain coherency (default=False)

    :return: freq,coherence,coherence_var,phase_coherence,phase_coherence_var,[coherency,coherency_t]: freq is an array
             of frequencies that the coherence was computed at. coherence is an array of length len(freq) that contains
             the coherence at each frequency. coherence_var is the variance of the coherence. phase_coherence is the
             average cosine phase difference at each frequency, and phase_coherence_var is the variance of that measure.
             coherency is only returned if return_coherency=True, it is the inverse fourier transform of the complex-
             valued coherency.
    """
    if s1.shape != s2.shape:
        raise AssertionError('s1 and s2 must have the same shape')
    if s1.ndim == 1:
        s1, s2 = [np.array(i, ndmin=2) for i in [s1, s2]]

    tf1, tf2 = list(), list()
    for i, (is1, is2) in enumerate(zip(s1, s2)):
        t1, freq1, itf1, rms1 = gaussian_stft(is1, sample_rate, window_length=window_length, increment=increment,
                                             min_freq=min_freq, max_freq=max_freq)

        t2, freq2, itf2, rms2 = gaussian_stft(is2, sample_rate, window_length=window_length, increment=increment,
                                             min_freq=min_freq, max_freq=max_freq)
        tf1.append(itf1)
        tf2.append(itf2)
    tf1, tf2 = [np.hstack(i) for i in [tf1, tf2]]

    cross_spec12 = tf1*np.conj(tf2)
    ps1 = np.abs(tf1)**2
    ps2 = np.abs(tf2)**2
    
    # compute the coherence using all the data
    csd = cross_spec12.sum(axis=1)
    denom = ps1.sum(axis=1)*ps2.sum(axis=1)
    c_amp = np.abs(csd) / np.sqrt(denom)
    cohe = c_amp**2

    # compute the phase coherence using all the data
    c_phase = np.cos(np.angle(csd))

    # make leave-one-out estimates of the complex coherence
    jn_estimates_amp = list()
    jn_estimates_phase = list()
    jn_estimates_cohe = list()
    
    njn = tf1.shape[1]
    for k in range(njn):
        i = np.ones([njn], dtype='bool')
        i[k] = False
        csd = cross_spec12[:, i].sum(axis=1)
        denom = ps1[:, i].sum(axis=1)*ps2[:, i].sum(axis=1)
        c_amp = np.abs(csd) / np.sqrt(denom)
        jn_estimates_amp.append(c_amp)
        jn_estimates_cohe.append(njn*cohe - (njn-1)*(c_amp**2))

        c_phase = np.cos(np.angle(csd))
        jn_estimates_phase.append(c_phase)
        

    jn_estimates_amp = np.array(jn_estimates_amp)
    jn_estimates_cohe = np.array(jn_estimates_cohe)
    jn_estimates_phase = np.array(jn_estimates_phase)

    # estimate the variance of the coherence
    jn_mean_amp = jn_estimates_amp.mean(axis=0)
    jn_diff_amp = (jn_estimates_amp - jn_mean_amp)**2
    c_var_amp = ((njn-1) / float(njn)) * jn_diff_amp.sum(axis=0)
    cohe_unbiased = jn_estimates_cohe.mean(axis=0)
    cohe_se = jn_estimates_cohe.std(axis=0)/np.sqrt(njn)
    
    
    # estimate the variance of the phase coherence
    jn_mean_phase = jn_estimates_phase.mean(axis=0)
    jn_diff_phase = (jn_estimates_phase - jn_mean_phase)**2
    c_phase_var = ((njn-1) / float(njn)) * jn_diff_phase.sum(axis=0)



    assert c_amp.max() <= 1.0, "c_amp.max()=%f" % c_amp.max()
    assert c_amp.min() >= 0.0, "c_amp.min()=%f" % c_amp.min()
    assert np.sum(np.isnan(c_amp)) == 0, "NaNs in c_amp!"

    if return_coherency:
        # compute the complex-valued coherency
        z = csd / denom

        # make the complex-valued coherency symmetric around zero
        sym_z = np.zeros([len(z)*2 - 1], dtype='complex')
        sym_z[len(z)-1:] = z
        sym_z[:len(z)-1] = (z[1:])[::-1]

        # do an fft shift so the inverse fourier transform works
        sym_z = fftshift(sym_z)

        if len(sym_z) % 2 == 1:
            # shift zero from end of shift_lags to beginning
            sym_z = np.roll(sym_z, 1)

        coherency = ifft(sym_z)
        coherency = fftshift(coherency.real)

        dt = 1. / sample_rate
        hc = (len(coherency) - 1) // 2
        coherency_t = np.arange(-hc, hc+1, 1)*dt

        """
        import matplotlib.pyplot as plt
        plt.figure()
        plt.plot(coherency_t, coherency, 'k-')
        plt.axis('tight')
        plt.show()
        """

        return freq1,c_amp,c_var_amp,c_phase,c_phase_var,coherency,coherency_t, cohe_unbiased, cohe_se
    else:
        return freq1,c_amp,c_var_amp,c_phase,c_phase_var, cohe_unbiased, cohe_se


def compute_coherence_from_timefreq(tf1, tf2, sample_rate, window_size, gauss_window=False, nstd=6):
    """ Compute the time-varying coherence between two complex-valued time-frequency representations.

    :param tf1: The first time-frequency representation.
    :param tf2: The second time-frequency representation.
    :param sample_rate: The temporal sample rate of the time-frequency representations (units=Hz)
    :param window_size: The size of the window used to average across samples for computing coherence (units=seconds)
    :param gauss_window: If True, use a gaussian weighting when averaging (default=False)
    :param nstd: The number of standard deviations wide the gaussian weighting is (default=6)

    :return: coherence: A array of shape (nfreqs, T) where nfreqs is the number of frequencies in the time-frequency
            representation, and T is the temporal length of the time-frequency representation.
    """

    N = tf1.shape[1]

    # compute the power spectrum of each individual spectrogram
    tf1_conj = np.conj(tf1)
    tf2_conj = np.conj(tf2)
    tf1_ps = (tf1 * tf1_conj).real
    tf2_ps = (tf2 * tf2_conj).real

    # compute the sufficient statistics for the cross spectrum, i.e. the stuff that will be averaged
    # when computing the coherence
    cross_spec12 = tf1 * tf2_conj
    cross_spec21 = tf1_conj * tf2
    del tf1_conj
    del tf2_conj

    # print('len(s1)=%d, sample_rate=%0.2f, increment=%0.6f, window_size=%0.3f' % (len(s1), self.sample_rate, increment, window_size))

    # nwinlen = max(np.unique(windows[:, 2] - windows[:, 1]))
    nwinlen = int(sample_rate*window_size)

    # print('nwindows=%d, nwinlen=%d' % (len(windows), nwinlen))
    # generate a normalized window for computing the weighted mean around a point in time
    if gauss_window:
        gauss_t, average_window = gaussian_window(nwinlen, nstd)
        average_window /= np.abs(average_window).sum()
    else:
        average_window = np.ones(nwinlen) / float(nwinlen)
        # print('len(average_window)=%d, average_window.sum()=%0.6f' % (len(average_window), average_window.sum()))

    nfreqs = tf1.shape[0]
    # compute the coherence at each frequency
    coherence = np.zeros([nfreqs, N])
    for k in range(nfreqs):
        # convolve the window function with each frequency band
        tf1_mean = convolve1d(tf1_ps[k, :], average_window, mode='mirror')
        tf2_mean = convolve1d(tf2_ps[k, :], average_window, mode='mirror')
        denom = tf1_mean * tf2_mean
        del tf1_mean
        del tf2_mean

        cs12_mean_r = convolve1d(cross_spec12[k, :].real, average_window, mode='mirror')
        cs12_mean_i = convolve1d(cross_spec12[k, :].imag, average_window, mode='mirror')
        cs12_mean = np.sqrt(cs12_mean_r**2 + cs12_mean_i**2)
        del cs12_mean_r
        del cs12_mean_i

        cs21_mean_r = convolve1d(cross_spec21[k, :].real, average_window, mode='mirror')
        cs21_mean_i = convolve1d(cross_spec21[k, :].imag, average_window, mode='mirror')
        cs21_mean = np.sqrt(cs21_mean_r**2 + cs21_mean_i**2)
        del cs21_mean_r
        del cs21_mean_i

        coherence[k, :] = (cs12_mean*cs21_mean) / denom

    return coherence


def mt_cross_coherence(s1, s2, sample_rate, window_size=5.0, increment=1.0, bandwidth=10.0, noise_floor=False, num_noise_trials=1, debug=False):
    """
        Compute the time-varying multi-taper cross coherence between two time series, with the option of computing a
        noise floor.

        s1,s2: the signals

        sample_rate: sample rate in Hz for the signal

        window_size: the size in seconds of the sliding window used to compute the coherence

        increment: the amount of time in seconds to slide the window forward per time point

        bandwidth: related to the number of tapers used to compute the cross spectral density

        noise_floor: whether or not to compute a lower bound on the coherence for each time point. The lower bound
            is defined by the average coherence between two signals that have the same power spectrum as s1 and s2
            but randomly shuffled phases.
    """

    assert len(s1) == len(s2)

    #compute lengths in # of samples
    nwinlen = int(sample_rate*window_size)
    if nwinlen % 2 == 0:
        nwinlen += 1
    hnwinlen = nwinlen // 2

    #compute increment in number of samples
    slen = len(s1)
    nincrement = int(sample_rate*increment)

    #compute number of windows
    nwindows = slen // nincrement

    #get frequency axis values by computing coherence between dummy slice
    win1 = np.zeros([nwinlen])
    win2 = np.zeros([nwinlen])
    cdata = compute_mtcoherence(win1+1.0, win2+1.0, sample_rate, window_size=window_size, bandwidth=bandwidth)
    freq = cdata.frequency

    #construct the time-frequency representation for time-varying coherence
    timefreq = np.zeros([len(freq), nwindows])
    if noise_floor:
        floor_window_index_min = int(np.ceil(hnwinlen / float(nincrement)))
        floor_window_index_max = nwindows - floor_window_index_min
        timefreq_floor = np.zeros([len(freq), nwindows])

    if debug:
        print('[cross_coherence] length=%0.3f, slen=%d, window_size=%0.3f, increment=%0.3f, bandwidth=%0.1f, nwindows=%d' %
              (slen/sample_rate, slen, window_size, increment, bandwidth, nwindows))

    #compute the coherence for each window
    #print('nwinlen=%d, hnwinlen=%d, nwindows=%d' % (nwinlen, hnwinlen, nwindows))
    for k in range(nwindows):
        if debug:
            stime = time.time()

        #get the indices of the window within the signals
        center = k*nincrement
        si = center - hnwinlen
        ei = center + hnwinlen + 1

        #adjust indices to deal with edge-padding
        sii = 0
        if si < 0:
            sii = abs(si)
            si = 0
        eii = sii + nwinlen
        if ei > slen:
            eii = sii + nwinlen - (ei - slen)
            ei = slen

        #set the content of the windows
        win1[:] = 0.0
        win2[:] = 0.0
        win1[sii:eii] = s1[si:ei]
        win2[sii:eii] = s2[si:ei]
        #print('(%0.2f, %0.2f, %0.2f), s1sum=%0.0f, s2sum=%0.0f, k=%d, center=%d, si=%d, ei=%d, sii=%d, eii=%d' % \)
        #      ((center-hnwinlen)/sample_rate, (center+hnwinlen+1)/sample_rate, center/sample_rate, s1sum, s2sum, k, center, si, ei, sii, eii)

        #compute the coherence
        cdata = compute_mtcoherence(win1, win2, sample_rate, window_size=window_size, bandwidth=bandwidth)
        timefreq[:, k] = cdata.coherence
        if debug:
            total_time = 0.0
            etime = time.time() - stime
            total_time += etime
            print('\twindow %d: time = %0.2fs' % (k, etime))

        #compute the noise floor
        if noise_floor:

            csum = np.zeros([len(cdata.coherence)])

            for m in range(num_noise_trials):
                if debug:
                    stime = time.time()

                #compute coherence between win1 and randomly selected slice of s2
                win2_shift_index = k
                while win2_shift_index == k or win2_shift_index < floor_window_index_min or win2_shift_index > floor_window_index_max:
                    win2_shift_index = np.random.randint(nwindows)
                w2center = win2_shift_index*nincrement
                w2si = w2center - hnwinlen
                w2ei = w2center + hnwinlen + 1
                win2_shift = s2[w2si:w2ei]
                #print('len(s2)=%d, win2_shift_index=%d, w2si=%d, w2ei=%d, len(win1)=%d, len(win2_shift)=%d' % \)
                #      (len(s2), win2_shift_index, w2si, w2ei, len(win1), len(win2_shift))
                cdata1 = compute_mtcoherence(win1, win2_shift, sample_rate, window_size=window_size, bandwidth=bandwidth)
                csum += cdata1.coherence

                #compute coherence between win2 and randomly selected slice of s1
                win1_shift_index = k
                while win1_shift_index == k or win1_shift_index < floor_window_index_min or win1_shift_index > floor_window_index_max:
                    win1_shift_index = np.random.randint(nwindows)
                w1center = win1_shift_index*nincrement
                w1si = w1center - hnwinlen
                w1ei = w1center + hnwinlen + 1
                win1_shift = s1[w1si:w1ei]
                #print('nwindows=%d, len(s1)=%d, win1_shift_index=%d, w1si=%d, w1ei=%d, len(win2)=%d, len(win1_shift)=%d' % \)
                #      (nwindows, len(s1), win1_shift_index, w1si, w1ei, len(win2), len(win1_shift))
                cdata2 = compute_mtcoherence(win2, win1_shift, sample_rate, window_size=window_size, bandwidth=bandwidth)
                csum += cdata2.coherence

                if debug:
                    etime = time.time() - stime
                    total_time += etime
                    print('\t\tnoise trial %d: time = %0.2fs' % (m, etime))

            timefreq_floor[:, k] = csum / (2.0*num_noise_trials)

        if debug:
            print('\tTotal time for window %d: %0.2fs' % (k, total_time))
            print('\tExpected total time for all iterations: %0.2f min' % (total_time*nwindows / 60.0))

    t = np.arange(nwindows)*increment
    if noise_floor:
        return t,freq,timefreq,timefreq_floor
    else:
        return t,freq,timefreq


def cross_spectral_density(s1, s2, sample_rate, window_length, increment, min_freq=0, max_freq=np.inf):
    """ Computes the cross-spectral density between signals s1 and s2 by computing the product of their power spectra.
        First the Gaussian-windowed spectrograms of s1 and s2 are computed, then the product of the power spectra
        at each time point is computed, then the products are averaged across time to produce an estimate of the
        cross spectral density.

    :param s1: The first signal.
    :param s2: The second signal.
    :param sample_rate: The sample rates of the signals.
    :param window_length: The length of the window used to compute the STFT (units=seconds)
    :param increment: The spacing between the points of the STFT  (units=seconds)
    :param min_freq: The minimum frequency to analyze (units=Hz, default=0)
    :param max_freq: The maximum frequency to analysize (units=Hz, default=nyquist frequency)

    :return: freq,csd: The frequency bands evaluated, and the cross-spectral density.
    """

    # compute the complex-spectrograms of the signals
    t1, freq1, tf1, rms1 = gaussian_stft(s1, sample_rate, window_length=window_length, increment=increment,
                                         min_freq=min_freq, max_freq=max_freq)

    t2, freq2, tf2, rms2 = gaussian_stft(s2, sample_rate, window_length=window_length, increment=increment,
                                         min_freq=min_freq, max_freq=max_freq)

    # multiply the complex spectrograms
    csd = tf1*tf2

    # take the power spectrum
    csd = np.abs(csd)

    # average across time
    csd = csd.mean(axis=1)

    return freq1,csd



from __future__ import division, print_function

import copy

import numpy as np
from scipy.interpolate import splrep,splev
from scipy.signal import hilbert
from scipy.stats import pearsonr

import matplotlib.pyplot as plt
from soundsig.signal import find_extrema


class IMF(object):
    """
        Container class for IMF and it's properties.
    """
    def __init__(self):
        self.imf = None
        self.std = None
        self.amplitude = None
        self.phase = None


class HHT(object):
    """
        An implementation of the Hilbert-Huang transform. Based on code from PyHHT:

        https://github.com/jaidevd/pyhht

        Three useful papers:

        N. E. Huang et al., "The empirical mode decomposition and the Hilbert spectrum for non-linear and non
        stationary time series analysis, Proc. Royal Soc. London A, Vol. 454, pp. 903-995, 1998

        Rato R.T., Ortigueira M.D., Batista A.G 2008 "On the HHT, its problems, and some solutions." Mechanical Systems
        and Signal Processing 22 1374-1394

        Huang, N. E., and Z. Wu (2008), A review on Hilbert-Huang transform: Method and its applications to geophysical studies,
        Rev. Geophys., 46, RG2006, doi:10.1029/2007RG000228
    """

    def __init__(self, s, sample_rate, emd_max_modes=30, emd_resid_tol=1e-3,
                 sift_mean_tol=1e-3, sift_stoppage_S=3, sift_max_iter=100, sift_remove_edge_effects=True,
                 ensemble_num_samples=1, ensemble_noise_gain=1.5,
                 hilbert_max_iter=20,
                 compute_on_init=True):
        """
            Initialize the Hilbert-Huang Transform (HHT). Parameters include:

            s: the signal to be transformed
            sample_rate: sample rate of the signal

            emd_max_modes: the maximum number of IMFs to extract (default=infinity)
            emd_resid_tol: stop EMD when the square sum of components of the residual is less than this quantity (default=1e-3)

            sift_mean_tol: stop sifting for an IMF when the sum of squares of the mean between upper and lower envelopes is lower than this value (default=1e-3)
            sift_stoppage_S: the number of iterations to apply the S-stoppage criteria (default=3)
            sift_max_iter: the maximum allowed number of sift iterations (default=100)
            sift_remove_edge_effects: whether or not to mirror the extrema on both ends of s to reduce edge effects (default=True)

            ensemble_num_samples: the number of samples to take when when doing ensemble EMD (default=1, implies no ensemble)
            ensemble_noise_gain: the number of standard deviations of the signal that are used to construct the standard deviation of the noise (default=1.5)

            hilbert_max_iter: maximum number of iterations when applying normalized Hilbert transform to an IMF

            compute_on_init: whether to perform the EMD when the constructor is called (default=True)
        """

        self.s = s
        self.sample_rate = sample_rate
        self.emd_max_modes = emd_max_modes
        self.emd_resid_tol = emd_resid_tol
        self.sift_mean_tol = sift_mean_tol
        self.sift_stoppage_S = sift_stoppage_S
        self.sift_max_iter = sift_max_iter
        self.sift_remove_edge_effects = sift_remove_edge_effects
        self.ensemble_num_samples = ensemble_num_samples
        self.ensemble_noise_gain = ensemble_noise_gain
        self.hilbert_max_iter = hilbert_max_iter

        self.imfs = list()

        if compute_on_init:
            self.compute_emd(self.s)

    def compute_imf_ensemble(self, s):
        """
            Computes the ensemble-empirical model decomposition (EEMD, Huang et. al 2008).
        """
        noise_std = self.ensemble_noise_gain*s.std()
        imf_mean = None
        imf_std = None

        for k in range(self.ensemble_num_samples):
            noise = np.random.randn(len(s))*noise_std
            imf = self.compute_imf(s+noise)
            if imf_mean is None:
                imf_mean = imf
                imf_std = np.zeros_like(imf_mean)
            else:
                resid = imf - imf_mean
                imf_mean = imf_mean + resid / (k+1)
                imf_std = imf_std + resid*(imf - imf_mean)

        imf_std = np.sqrt(imf_std) / (self.ensemble_num_samples - 1)

        return imf_mean,imf_std

    def compute_imf(self, s, plot=False):
        """
            Compute an intrinsic mode function from a signal s using sifting.
        """

        stop = False
        #make a copy of the signal
        imf = copy.copy(s)
        #find extrema for first iteration
        mini,maxi = find_extrema(s)

        if len(mini) == 0 or len(maxi) == 0:
            return None

        #keep track of extrema difference
        num_extrema = np.zeros([self.sift_stoppage_S, 2])  # first column are maxima, second column are minima
        num_extrema[-1, :] = [len(maxi), len(mini)]
        iter = 0
        while not stop:

            #set some things up for the iteration
            s_used = s
            left_padding = 0
            right_padding = 0

            #add an extra oscillation at the beginning and end of the signal to reduce edge effects; from Rato et. al (2008) section 3.2.2
            if self.sift_remove_edge_effects:
                Tl = maxi[0]  # index of left-hand (first) maximum
                tl = mini[0]  # index of left-hand (first) minimum

                Tr = maxi[-1]  # index of right hand (last) maximum
                tr = mini[-1]  # index of right hand (last) minimum

                #to reduce end effects, we need to extend the signal on both sides and reflect the first and last extrema
                #so that interpolation works better at the edges
                left_padding = max(Tl, tl)
                right_padding = len(s) - min(Tr, tr)

                #pad the original signal with zeros and reflected extrema
                s_used = np.zeros([len(s) + left_padding + right_padding])
                s_used[left_padding:-right_padding] = s

                #reflect the maximum on the left side
                imax_left = left_padding-tl
                s_used[imax_left] = s[Tl]
                #reflect the minimum on the left side
                imin_left = left_padding-Tl
                s_used[imin_left] = s[tl]

                #correct the indices on the right hand side so they're useful
                trr = len(s) - tr
                Trr = len(s) - Tr

                #reflect the maximum on the right side
                roffset = left_padding + len(s)
                imax_right = roffset+trr-1
                s_used[imax_right] = s[Tr]
                #reflect the minimum on the right side
                imin_right = roffset+Trr-1
                s_used[imin_right] = s[tr]

                #extend the array of maxima
                new_maxi = [i + left_padding for i in maxi]
                new_maxi.insert(0, imax_left)
                new_maxi.append(imax_right)
                maxi = new_maxi

                #extend the array of minima
                new_mini = [i + left_padding for i in mini]
                new_mini.insert(0, imin_left)
                new_mini.append(imin_right)
                mini = new_mini

            t = np.arange(0, len(s_used))
            fit_index = range(left_padding, len(s_used)-right_padding)

            #fit minimums with cubic splines
            spline_order = 3
            if len(mini) <= 3:
                spline_order = 1
            min_spline = splrep(mini, s_used[mini], k=spline_order)
            min_fit = splev(t[fit_index], min_spline)

            #fit maximums with cubic splines
            spline_order = 3
            if len(maxi) <= 3:
                spline_order = 1
            max_spline = splrep(maxi, s_used[maxi], k=spline_order)
            max_fit = splev(t[fit_index], max_spline)

            if plot:
                plt.figure()
                plt.plot(t[fit_index], max_fit, 'r-')
                plt.plot(maxi, s_used[maxi], 'ro')
                plt.plot(left_padding, 0.0, 'kx', markersize=10.0)
                plt.plot(left_padding+len(s), 0.0, 'kx', markersize=10.0)
                plt.plot(t, s_used, 'k-')
                plt.plot(t[fit_index], min_fit, 'b-')
                plt.plot(mini, s_used[mini], 'bo')
                plt.suptitle('Iteration %d' % iter)

            #take average of max and min splines
            z = (max_fit + min_fit) / 2.0

            #compute a factor used to dampen the subtraction of the mean spline; Rato et. al 2008, sec 3.2.3
            alpha,palpha = pearsonr(imf, z)
            alpha = min(alpha, 1e-2)

            #subtract off average of the two splines
            d = imf - alpha*z

            #set the IMF to the residual for next iteration
            imf = d

            #check for IMF S-stoppage criteria
            mini,maxi = find_extrema(imf)
            num_extrema = np.roll(num_extrema, -1, axis=0)
            num_extrema[-1, :] = [len(mini), len(maxi)]
            if iter >= self.sift_stoppage_S:
                num_extrema_change = np.diff(num_extrema, axis=0)
                de = np.abs(num_extrema[-1, 0] - num_extrema[-1, 1])
                if np.abs(num_extrema_change).sum() == 0 and de < 2 and np.abs(imf.mean()) < self.sift_mean_tol:
                    stop = True
            if iter > self.sift_max_iter:
                stop = True
            print('Iter %d: len(mini)=%d, len(maxi=%d), imf.mean()=%0.6f, alpha=%0.2f' % (iter, len(mini), len(maxi), imf.mean(), alpha))
            #print('num_extrema=',num_extrema)
            iter += 1
        return imf

    def compute_emd(self, s):
        """
            Perform the empirical mode decomposition on a signal s.
        """

        self.imfs = list()
        #make a copy of the signal that will hold the residual
        r = copy.copy(s)
        stop = False
        while not stop:
            #compute the IMF from the signal
            if self.ensemble_num_samples == 1:
                imf_mean = self.compute_imf(r)
                if imf_mean is None:
                    stop = True
                    break
                imf_std = np.zeros_like(imf_mean)
            else:
                imf_mean,imf_std = self.compute_imf_ensemble(r)

            #compute the normalized hilbert transform
            #am,fm,phase,ifreq = self.normalized_hilbert(imf_mean)
            amplitude,phase = self.hilbert(imf_mean)

            #construct an IMF object
            imf = IMF()
            imf.imf = imf_mean
            imf.std = imf_std
            imf.amplitude = amplitude
            imf.phase = phase

            self.imfs.append(imf)

            #subtract the IMF off to produce a new residual
            r -= imf_mean

            #compute extrema for detecting a trend IMF
            maxi,mini = find_extrema(r)

            #compute convergence criteria
            if np.abs(r).sum() < self.emd_resid_tol or len(self.imfs) == self.emd_max_modes or (len(maxi) == 0 and len(mini) == 0):
                stop = True

        #append the residual as the last mode
        self.emd_residual = r

    def hilbert(self, s):
        """
            Perform the Hilbert transform on the signal s.
        """
        z = hilbert(s)
        return np.abs(z),np.angle(z)


    def normalized_hilbert(self, s):
        """
            Perform the "Normalized" Hilbert transform (Huang 2008 sec. 3.1) on the IMF s, decomposing the
            signal s into AM and FM components. Returns am,fm,phase,ifreq - am is the AM component, fm is the FM
            component, phase is the the arccos of fm, and ifreq is the instantaneous frequency.
        """

        x = copy.copy(s)
        iter = 0
        converged = False

        while not converged:
            #take the absolute value of the IMF and find the extrema
            absx = np.abs(x)
            mini,maxi = find_extrema(absx)
            if len(mini) == 0 or len(maxi) == 0:
                converged = True
                break
            spline_order = 3

            #reflect first and last maxima to remove edge effects for interpolation
            left_padding = maxi[0]
            right_padding = len(x) - maxi[-1]
            x_padded = np.zeros([len(x)+left_padding+right_padding])
            x_padded[left_padding:-right_padding] = absx
            x_padded[0] = absx[maxi[0]]
            x_padded[-1] = absx[maxi[-1]]

            #create new array of extremas
            new_maxi = [i+left_padding for i in maxi]
            new_maxi.insert(0, 0)
            new_maxi.append(len(x_padded)-1)

            #fit a cubic spline to the extrema of the absolute value of the signal
            if len(maxi) <= 3:
                spline_order = 1
            max_spline = splrep(new_maxi, x_padded[new_maxi], k=spline_order)

            #use the spline to interpolate over the course of the signal
            t = np.arange(0, len(x_padded))
            fit_index = range(left_padding, len(x_padded)-right_padding)
            env = splev(t[fit_index], max_spline)

            """
            plt.figure()
            plt.plot(t, x_padded, 'k-')
            plt.plot(new_maxi, x_padded[new_maxi], 'ro-', markersize=8)
            plt.axis('tight')
            plt.suptitle('Iter %d' % iter)
            """

            #divide by envelope
            x /= env

            #check for convergence
            iter += 1
            if iter >= self.hilbert_max_iter:
                converged = True
            if (x.max() - 1.0) <= 1e-6:
                converged = True
            if len(mini) == 0 or len(maxi) == 0:
                converged = True

        #compute the FM and AM components
        fm = x
        am = s / fm

        #compute the phase
        phase = np.arccos(fm)

        #compute the instantaneous frequency
        ifreq = np.zeros([len(phase)])
        ifreq[1:] = np.diff(phase) * self.sample_rate

        return am,fm,phase,ifreq



from __future__ import division, print_function

import operator

import numpy as np
from scipy.stats import gamma
from scipy.ndimage import convolve1d

import matplotlib.pyplot as plt
from matplotlib.colors import ListedColormap
from matplotlib.patches import Rectangle


def compute_joint_isi(spike_train1, spike_train2, window_size=0.500, bin_size=0.001):

    half_window_size = window_size / 2.0
    half_nbins = int(half_window_size / bin_size)

    nbins = half_nbins*2 + 1 # ensure an odd number of bins, with zero lag in the middle

    #construct sparse matrix of spike-to-spike distances
    isi_hist = np.zeros([nbins], dtype='int')

    lowest_j = 0
    for i,ti in enumerate(spike_train1):
        #print 'i=%d, ti=%0.3f, lowest_j=%d' % (i, ti, lowest_j)
        if lowest_j > len(spike_train2)-1:
            break
        for j in range(lowest_j, len(spike_train2)):
            tj = spike_train2[j]
            dij = ti - tj
            #print '\tj=%d, tj=%0.3f, dij=%0.3f' % (j, tj, dij)

            if dij > half_window_size:
                #there is no t{i+k}, k>0 such that t{i+k} - tj < half_window_size, so this is the lowest that
                # j has to be for future iterations. we'll keep track of that to reduce the number of iterations
                # of the inner loop for future outer loop iterations
                lowest_j = j+1
                continue

            if dij < -half_window_size:
                #at this point there is no tj such that ti - tj >= -half_window_size, so we should break
                break

            else:
                #add to the histogram
                bin_index = int(np.round(dij / bin_size)) + half_nbins
                #print '\t  added to bin, bin_index=%d' % bin_index
                isi_hist[bin_index] += 1

    sp = window_size / nbins
    isi_vals = np.arange(-half_window_size, half_window_size, sp)   #values of left hand edges of bins
    return isi_vals,isi_hist


def simulate_poisson(psth, duration, num_trials=20):

    dt = 0.001
    trange = np.arange(0.0, duration, dt)
    new_spike_trials = []
    for k in range(num_trials):
        next_spike_time = np.random.exponential(1.0)
        last_spike_index = 0
        spike_times = []
        for k, t in enumerate(trange):
            csum = np.cumsum(psth[last_spike_index:k])
            if len(csum) < 1:
                continue
            if csum[-1] >= next_spike_time:
                last_spike_index = k
                spike_times.append(t)
        new_spike_trials.append(np.array(spike_times))
    return new_spike_trials


def simulate_gamma(psth, trials, duration, num_trials=20):

    #rescale the ISIs
    dt = 0.001
    rs_isis = []
    for trial in trials:
        if len(trial) < 1:
            continue
        csum = np.cumsum(psth)*dt
        for k,ti in enumerate(trial[1:]):
            tj = trial[k]
            if ti > duration or tj > duration or ti < 0.0 or tj < 0.0:
                continue
            ti_index = int((ti / duration) * len(psth))
            tj_index = int((tj / duration) * len(psth))
            #print 'k=%d, ti=%0.6f, tj=%0.6f, duration=%0.3f' % (k, ti, tj, duration)
            #print '  ti_index=%d, tj_index=%d, len(psth)=%d, len(csum)=%d' % (ti_index, tj_index, len(psth), len(csum))
            #get rescaled time as difference in cumulative intensity
            ui = csum[ti_index] - csum[tj_index]
            if ui < 0.0:
                print('ui < 0! ui=%0.6f, csum[ti]=%0.6f, csum[tj]=%0.6f' % (ui, csum[ti_index], csum[tj_index]))
            else:
                rs_isis.append(ui)
    rs_isis = np.array(rs_isis)
    rs_isi_x = np.arange(rs_isis.min(), rs_isis.max(), 1e-5)

    #fit a gamma distribution to the rescaled ISIs
    gamma_alpha,gamma_loc,gamma_beta = gamma.fit(rs_isis)
    gamma_pdf = gamma.pdf(rs_isi_x, gamma_alpha, loc=gamma_loc, scale=gamma_beta)
    print('Rescaled ISI Gamma Fit Params: alpha=%0.3f, beta=%0.3f, loc=%0.3f' % (gamma_alpha, gamma_beta, gamma_loc))

    #simulate new trials using rescaled ISIs
    new_trials = []
    for nt in range(num_trials):
        ntrial = []
        next_rs_time = gamma.rvs(gamma_alpha, loc=gamma_loc,scale=gamma_beta)
        csum = 0.0
        for t_index,pval in enumerate(psth):
            csum += pval*dt
            if csum >= next_rs_time:
                #spike!
                t = t_index*dt
                ntrial.append(t)
                #reset integral and generate new rescaled ISI
                csum = 0.0
                next_rs_time = gamma.rvs(gamma_alpha, loc=gamma_loc,scale=gamma_beta)
        new_trials.append(ntrial)
    #plt.figure()
    #plt.hist(rs_isis, bins=20, normed=True)
    #plt.plot(rs_isi_x, gamma_pdf, 'r-')
    #plt.title('Rescaled ISIs')

    return new_trials


def compute_psth(trials, duration, bin_size=0.001, time_offset=0.0):
    """
        Compute a peri-stimulus time histogram (PSTH), conditioned on an event such as stimulus.

        trials: an array of arrays of spike times in seconds, relative to the onset of the stimulus,
                If a spike precedes a stimulus, it's spike time should be negative. len(trials) = # of trials,
                and len(trials[0]) = number of spikes in first trial
        duration: the duration of the event.
        bin_size: the size in seconds of the bin to use in creating the PSTH (defaults to 0.001s = 1ms)

        Returns the average spike rate in KHz across trials in each time bin.
    """

    nbins = int(np.ceil((duration) / bin_size))
    spike_counts = np.zeros(nbins)
    for stimes in trials:
        if len(stimes) == 0:
            continue
        stimes = np.array(stimes)
        if len(stimes.shape) > 0:
            # get index of spike times valid for the conditioned event
            vi = (stimes >= time_offset) & (stimes <= duration)

            # convert spike times to indices in PSTH
            sbins = np.floor((stimes[vi]-time_offset) / bin_size).astype('int')

            # add spike to each bin
            for k in sbins:
                if k < nbins:
                    spike_counts[k] += 1

    # compute rate in KHz by dividing by bin size
    spike_counts /= bin_size*1000.0

    # take mean across trials (spikes were already summed across trials)
    spike_counts /= len(trials)

    # construct time axis, represents time point at left hand of bin
    t = (np.arange(nbins).astype('float') * bin_size) + time_offset

    return t,spike_counts


def create_random_psth(duration, smooth_win_size=10, samp_rate=1000.0, thresh=0.5):
    nsamps = duration * samp_rate
    psth = np.random.randn(nsamps)
    psth[psth < thresh] = 0.0

    #smooth psth
    kt = np.arange(-smooth_win_size, smooth_win_size+1, 1.0)
    k = np.exp(-kt**2)
    k /= k.sum()
    psth = np.convolve(psth, k, mode='same')

    return psth


def plot_raster(spike_trains, ax=None, duration=None, bin_size=0.001, time_offset=0.0, ylabel='Trial #', groups=None,
                bgcolor=None, spike_color='k'):
    """
        Make a raster plot of the trials of spike times.

        spike_trains: an array of arrays of spike times in seconds.
        time_offset: amount of time in seconds to offset the time axis for plotting
        groups: a dictionary that groups spike trains together. the key is the group name, and
            the value is a list of spike train indicies. The groups are
            differentiated visually using a background color, and labeled on the y-axis.
            The elements in the indicies array must be contiguous!
    """

    if ax is None:
        ax = plt.gca()

    if bgcolor is not None:
        ax.set_axis_bgcolor(bgcolor)

    if duration is None:
        duration = -np.inf
        for trial in spike_trains:
            if len(trial) > 0:
                duration = max(duration, np.max(trial))

    nbins = (duration / bin_size)

    #draw group backgrounds
    if groups is not None:

        #first make sure indicies are lists
        groups = dict([(kk, vv if type(vv) is list else [vv]) for kk, vv in groups.iteritems()])
        #sort group names by min trial        
        group_list = [(group_name,min(trial_indicies)) for group_name,trial_indicies in groups.iteritems()]
        group_list.sort(key=operator.itemgetter(1))
        group_list = [x[0] for x in group_list]
        for k,(group_name,trial_indicies) in enumerate(groups.iteritems()):
            real_index = group_list.index(group_name)
            if real_index % 2:
                max_trial = max(trial_indicies)
                y = len(spike_trains) - max_trial - 1
                x = 0.0 + time_offset
                h = len(trial_indicies)
                w = nbins
                rect = Rectangle( (x, y), width=w, height=h, fill=True, alpha=0.5, facecolor='#aaaaaa', linewidth=0.0)
                ax.add_patch(rect)

    #draw actual spikes
    for k,trial in enumerate(spike_trains):
        if len(trial) == 0:
            continue
        for st in trial:
            y = len(spike_trains) - k - 1
            x = st
            rect = Rectangle( (x, y), width=bin_size, height=1, linewidth=1.0, facecolor=spike_color, edgecolor=spike_color)
            ax.add_patch(rect)

    #change x axis tick marks to reflect actual time
    ax.autoscale_view()
    ax.set_xlim(time_offset, time_offset+duration)
    ax.figure.canvas.draw()

    if groups is None:
        #change y axis tick labels to reflect trial number
        y_labels = [y.get_text() for y in ax.get_yticklabels()]
        y_labels.reverse()
        ax.set_yticklabels(y_labels)
    else:
        ax.set_yticklabels([])
        #change y axis tick labels to reflect group, one tick per group
        yticks = list()
        for k,(group_name,trial_indicies) in enumerate(groups.iteritems()):
            min_trial = min(trial_indicies)
            ypos = len(spike_trains) - (min_trial + (len(trial_indicies) / 2.0))
            yticks.append( (ypos, group_name) )
        yticks.sort(key=operator.itemgetter(0))
        ax.set_yticks([y[0] for y in yticks])
        ax.set_yticklabels([y[1] for y in yticks])

    if ylabel is not None:
        ax.set_ylabel(ylabel)
    ax.set_xlabel('Time (s)')


def xcorr_hist(spike_train1, spike_train2, duration=None, window_size=0.001, sample_rate=1000.0, normalize=True):
    """
        Make a cross-correlation histogram of coincident spike times between spike train 1 and 2. The cross-correlation
        histogram is a function of time. At each moment t im time, the value of the histogram is given as the number
        of spike pairs from train 1 and 2 that are within the window specified by window_size.

        Normalization means to divide by window_size*int(duration*sample_rate), which turns the returned quantity into
        the probability of spikes from the two trains co-occurring.

        Returns t,xhist,clow,chigh where t is the time vector, xhist is the cross-correlation histogram, and clow and chigh
        are the lower and upper 95% confidence intervals. When a normalized xhist falls between these
    """

    if duration is None:
        duration = -np.inf
        for st in spike_train1:
            if len(st) > 0:
                duration = np.max(np.max(st), duration)

    #construct the histogram
    nbins = int(np.ceil(duration*sample_rate))
    xhist = np.zeros([nbins], dtype='int')

    half_window_size = window_size / 2
    #populate the histogram
    for t in range(nbins):
        tmin = t/sample_rate - half_window_size
        tmax = t/sample_rate + half_window_size
        #count the number of spikes that occur in this time window
        ns1 = ((spike_train1 >= tmin) & (spike_train1 <= tmax)).sum()
        ns2 = ((spike_train2 >= tmin) & (spike_train2 <= tmax)).sum()
        #compute the count of all pairs, this is the value for the histogram
        xhist[t] = ns1*ns2

    R = int(duration*sample_rate)
    if normalize:
        xhist = xhist.astype('float') / (window_size * R)

    #compute confidence intervals
    clow = -1.96 / np.sqrt(4*window_size*R)
    chigh = 1.96 / np.sqrt(4*window_size*R)

    t = np.arange(nbins)*(1.0 / sample_rate)
    return t,xhist,clow,chigh


def spike_envelope(spike_trains, start_time, duration, bin_size=1e-3, win_size=3.0, thresh_percentile=None, smooth=False):

    #construct empty envelope
    tlen = int(duration / bin_size)
    env = np.zeros([tlen])

    #sum spike trains across electrodes
    for st in spike_trains:

        #some basic checks
        assert np.sum(st < start_time) == 0, "spike_envelope: %d spike times occurred before the start time of %0.6fs" % (np.sum(st < start_time), start_time)
        assert np.sum(st > start_time+duration) == 0, "spike_envelope: %d spike times occurred after the end time of %0.6fs" % (np.sum(st > start_time+duration), start_time+duration)

        #convert spike times to indices
        sindex = ((st - start_time) / bin_size).astype('int')

        #increment spike count vector
        env[sindex] += 1

    if smooth:
        #smooth the spike count vector with a gaussian
        sct = np.linspace(-50, 50, 30)
        scwin = np.exp(-(sct**2 / win_size**2))
        env = convolve1d(env, scwin)

    #normalize the envelope
    env /= env.max()

    assert np.sum(env < 0.0) == 0, "Why are there zeros in the spike envelope?"

    if thresh_percentile is not None:
        thresh = np.percentile(env, thresh_percentile)
        print('spike_envelope threshold: %f' % thresh)
        env[env < thresh] = 0.0

    return env


def spike_trains_to_matrix(spike_trains, bin_size, start_time, duration):
    """ Convert an array of spike time arrays to a matrix of counts.

    :param spike_trains: An array of arrays of spike times.
    :param bin_size: The bin size of each matrix pixel.
    :param start_time: The start time of the matrix.
    :param duration: The duration of the matrix.
    :return: A matrix of spike counts, one row per each array in the spike_trains array.
    """

    nt = int(duration / bin_size)
    spike_count = np.zeros([len(spike_trains), nt])
    for k, spikes in enumerate(spike_trains):
        vi = (spikes >= start_time) & (spikes < start_time+duration)
        # convert the spike times into integer indices in spike_count
        spikes_index = ((spikes[vi] - start_time) / bin_size).astype('int')
        #increment each bin by the number of spikes that lie in it
        for si in spikes_index:
            assert si <= spike_count.shape[1], "IndexError: nt=%d, si=%d, k=%d" % (nt, si, k)
            spike_count[k, min(si, spike_count.shape[1]-1)] += 1.0
    return spike_count


def psth_colormap(noise_level=0.1, ncolors=256):

    cdata = list()
    for x in np.linspace(0, 1, ncolors):

        if x < noise_level:
            cdata.append([1., 1., 1., 1])
        else:
            v = (x - noise_level) / (1. - noise_level)
            c = (1. - v)**6
            # cdata.append([0, v/2., v, (v/2. + 0.5)])
            cdata.append([c, c, c])

    return ListedColormap(cdata, name='psth')


def causal_smooth(spike_times, duration, bin_size=1e-3, tau=1e-3, winlen=5e-3, num_win_points=11):
    """ Convolve a set of spike times (specified in seconds) with a causal exponential
        filter with time constant tau.
    """

    assert num_win_points % 2 == 1

    # convert the spike times to a binary vector
    nbins = int(duration / bin_size)
    b = np.zeros(nbins)
    sti = (spike_times / bin_size).astype('int')
    sti[sti < 0] = 0
    sti[sti > nbins-1] = nbins-1
    b[sti] = 1

    # create an causal exponential window
    x = np.linspace(-winlen, winlen, num_win_points)
    w = np.exp(-x / tau)
    w[x < 0] = 0
    w /= w.sum()

    return convolve1d(b, w)


def simple_synchrony(spike_times1, spike_times2, duration, bin_size=1e-1):
    """ Turn the two spike trains into binary vectors by binning, compute their normalized distance. Should
        be bounded by 0 and 1. """

    nbins = int(duration / bin_size)
    b1 = np.zeros(nbins, dtype='bool')
    b2 = np.zeros(nbins, dtype='bool')

    sti1 = (spike_times1 / bin_size).astype('int')
    sti1[sti1 < 0] = 0
    sti1[sti1 > nbins - 1] = nbins - 1
    b1[sti1] = True

    sti2 = (spike_times2 / bin_size).astype('int')
    sti2[sti2 < 0] = 0
    sti2[sti2 > nbins - 1] = nbins - 1
    b2[sti2] = True

    n1 = b1.sum()
    n2 = b2.sum()

    return np.sum(b1 & b2) / np.sqrt(n1*n2)


def exp_conv(spike_times, duration, tau, bin_size, causal=True):
    """ Convolve spike train with an exponential kernel. 
    
    :param spike_times: List of spike times in seconds.
    :param tau: Exponential time constant in seconds.
    :param duration: The duration of the time series.
    :param bin_size: Bin size in seconds
    :param causal: Whether to use a causal filter or not. If causal=False, then the spike times are convolved with a two-sided exponential
    
    :return: An array time series. 
    """

    assert spike_times.min() >= 0, "No negative spike times for exp_conv!"

    nt = int(duration / bin_size)

    good_spikes = (spike_times > 0) & (spike_times < duration)
    i = (spike_times[good_spikes] / bin_size).astype('int')

    s = np.zeros([nt])
    s[i] = 1.

    # make sure the exponential window size is at least 4 times the time constant
    winlen = 4*int(tau/bin_size) + 1
    assert winlen < len(s), "Signal is too small to convolve with exponential that has tau=%0.3f" % tau
    hwinlen = int(winlen / 2)
    twin = np.arange(-hwinlen, hwinlen+1)*bin_size
    win = np.zeros_like(twin)
    win[hwinlen:] = np.exp(-twin[hwinlen:] / tau)
    if ~causal:
        win[:hwinlen] = win[(hwinlen+1):][::-1]

    sc = convolve1d(s, win)

    return sc



from __future__ import division, print_function

import numpy as np


def compute_R2(x, y):
    return(((x - x.mean()) * (y - y.mean())).mean() / (x.std()*y.std()))**2


def effective_dof(X, lambda_val):
    """ Compute the effective degrees-of-freedom for data matrix X and lambda value. """
    s = np.linalg.svd(X, compute_uv=False)
    return np.sum(s**2 / (s**2 + lambda_val))



from __future__ import print_function

import importlib
import operator

packages = ['numpy', 'scipy', 'matplotlib', 'h5py', 'sklearn', 'networkx',
            'cython', 'numexpr', 'tables', 'nitime', 'joblib', 'spams']
missing = list()
versions = dict()

for pkg in packages:
    try:
        mod = importlib.import_module(pkg)
        version = 'n/a'
        if hasattr(mod, '__version__'):
            version = getattr(mod, '__version__')
        versions[pkg] = version
    except ImportError:
        missing.append(pkg)

print('Missing Packages: %s' % ','.join(missing))
print('')
print('Installed Packages:')

for pkg,version in sorted(versions.items(), key=operator.itemgetter(0)):
    print('%s: %s' % (pkg, version))

